import logging
import io
import numbers
from collections import Counter

import numpy as np

logging.getLogger('hdf5plugin').addHandler(logging.NullHandler())

import h5py
try:
    import hdf5plugin
except ImportError:
    hdf5plugin = None

from ..mzml.binary_encoding import encode_array_direct, encoding_map, compression_map, dtype_to_encoding, COMPRESSION_NONE
from ..mzml.writer import PlainMzMLWriter as _MzMLWriter, NON_STANDARD_ARRAY, ARRAY_TYPES, Mapping
from ..mzml.index import IndexingStream

from . import components


HDF5_COMPRESSORS = {}

DEFAULT_COMPRESSOR = 'gzip'

if hdf5plugin is not None:
    HDF5_COMPRESSORS = {
        "blosc": hdf5plugin.Blosc(),
        "blosc:lz4": hdf5plugin.Blosc('lz4'),
        "blosc:lz4hc": hdf5plugin.Blosc('lz4hc'),
        "blosc:zlib": hdf5plugin.Blosc('zlib'),
        "blosc:zstd": hdf5plugin.Blosc('zstd'),
    }
    HDF5_COMPRESSORS = {k: dict(v) for k, v in HDF5_COMPRESSORS.items()}
    DEFAULT_COMPRESSOR = 'blosc:zstd'

HDF5_COMPRESSORS['zlib'] = HDF5_COMPRESSORS['gzip'] = {'compression': 'gzip', 'compression_opts': 4}


HDF5_COMPRESSOR_MAGIC_NUMBERS_TO_NAME = {
    v['compression']: k for k, v in HDF5_COMPRESSORS.items()
}


class MzMLbWriter(_MzMLWriter):
    '''A high level API for generating mzMLb HDF5 files from simple Python objects.

    This class's public interface is identical to :class:`~.MzMLWriter`, with the exception of those
    related to HDF5 compression described below.

    Attributes
    ----------
    h5_compression : str
        A valid HDF5 compressor ID or compression scheme name or :const:`None`. Available compression schemes
        are "gzip"/"zlib", and if :mod:`hdf5plugin` is installed, "blosc", "blosc:lz4", "blosc:zlib", and
        "blosc:zstd". All Blosc-based compressors enable byte shuffling.
    h5_compressor_options : int or tuple
        The options to provide to the compressor designated by :attr:`h5_compressor`. For "gzip", this a single
        integer setting the compression level, while Blosc takes a tuple of integers.
    h5_blocksize : int
        The number of bytes to include in a single HDF5 data block. Smaller blocks improve random access speed
        at the expense of compression efficiency and space. Defaults to 2 ** 20, 1MB.
    '''
    def __init__(self, h5_file, close=False, vocabularies=None, missing_reference_is_error=False,
                 vocabulary_resolver=None, id=None, accession=None, h5_compression=DEFAULT_COMPRESSOR,
                 h5_compression_options=None, h5_blocksize=2**20, **kwargs):
        if h5_compression in HDF5_COMPRESSORS:
            key = h5_compression
            h5_compression = HDF5_COMPRESSORS[key]['compression']
            if h5_compression_options is None:
                h5_compression_options = HDF5_COMPRESSORS[key]['compression_opts']
        if h5_compression_options is None:
            h5_compression_options = 4
        self.compressor_name = HDF5_COMPRESSOR_MAGIC_NUMBERS_TO_NAME[h5_compression]
        if not isinstance(h5_file, h5py.File):
            h5_file = h5py.File(h5_file, 'w')
        self.xml_buffer = io.BytesIO()
        outfile = IndexingStream(self.xml_buffer)
        super(MzMLbWriter, self).__init__(
            outfile, close, vocabularies, missing_reference_is_error, vocabulary_resolver,
            id, accession, **kwargs)
        self.index_builder = outfile

        self.h5_file = h5_file
        self.h5_blocksize = h5_blocksize
        self.h5_compression = h5_compression
        self.h5_compression_options = h5_compression_options

        self.array_name_cache = {}
        self.offset_tracker = Counter()

    def begin(self):
        self.h5_file.attrs['compression'] = self.compressor_name
        return super(MzMLbWriter, self).begin()

    def end(self, type, value, traceback):
        close_ = self._close
        self._close = False
        super(MzMLbWriter, self).end(type, value, traceback)
        xml_bytes = self.xml_buffer.getvalue()
        n = len(xml_bytes)
        self.h5_file.create_dataset(
            'mzML', shape=(n, ), chunks=True,
            dtype=np.int8, data=bytearray(xml_bytes), compression=self.h5_compression,
            compression_opts=self.h5_compression_options)
        self.h5_file['mzML'].attrs['version'] = "mzMLb 1.0"
        for array, z in self.offset_tracker.items():
            self.h5_file[array].resize((z, ))

        for index in self.index_builder.indices:
            self._prepare_index(index, index.name, n)

        self._close = close_
        self.h5_file.flush()
        if self._should_close():
            self.close()

    def _generate_array_name(self, array_type, is_non_standard=False, scope='spectrum', dtype=None):
        if not is_non_standard:
            cv_ref, name, accession, term = self.context._resolve_cv_ref(array_type, None, None)
            key = accession.replace(":", "_")
        else:
            key = array_type.replace(" ", "_")
        tag_name = "{scope}_{key}".format(scope=scope, key=key)
        if dtype is not None:
            tag_name += '_' + dtype.__name__
        self.h5_file.create_dataset(
            tag_name, chunks=(self.h5_blocksize, ),
            shape=(self.h5_blocksize, ), dtype=dtype, compression=self.h5_compression,
            compression_opts=self.h5_compression_options, maxshape=(None, ))
        return tag_name

    def _prepare_array(self, array, encoding=32, compression=COMPRESSION_NONE, array_type=None,
                       default_array_length=None, scope='spectrum'):
        if isinstance(encoding, numbers.Number):
            _encoding = int(encoding)
        else:
            _encoding = encoding
        dtype = encoding_map[_encoding]
        array = np.array(array, dtype=dtype)
        encoded_array = encode_array_direct(
            array, compression=compression, dtype=dtype)

        if default_array_length is not None and len(array) != default_array_length:
            override_length = True
        else:
            override_length = False
        is_non_standard = False
        params = []
        if array_type is not None:
            params.append(array_type)
            if isinstance(array_type, Mapping):
                array_type_ = array_type['name']
            else:
                array_type_ = array_type
            if array_type_ not in ARRAY_TYPES:
                is_non_standard = True
                params.append(NON_STANDARD_ARRAY)
        params.append(compression_map[compression])
        params.append(dtype_to_encoding[dtype])

        if isinstance(array_type, dict):
            if len(array_type) == 1:
                array_name = next(iter(array_type.keys()))
            else:
                array_name = array_type['name']
        else:
            array_name = array_type
        try:
            storage_name = self.array_name_cache[array_name, is_non_standard, scope, dtype]
        except KeyError:
            storage_name = self._generate_array_name(array_name, is_non_standard, scope, dtype)
            self.array_name_cache[array_name, is_non_standard, scope, dtype] = storage_name

        length = len(encoded_array)
        offset = self.offset_tracker[storage_name]

        buff = self.h5_file[storage_name]
        buff.resize((offset + length, ))
        buff[offset:offset + length] = encoded_array

        self.offset_tracker[storage_name] += length

        return self.ExternalBinaryDataArray(
            external_dataset_name=storage_name,
            offset=offset, array_length=length,
            params=params)

    def _prepare_index(self, index, name, last):
        offset_index_key = "mzML_{name}Index".format(name=name)
        offset_index_id_ref = "mzML_{name}Index_idRef".format(name=name)

        id_ref_array = []
        offset_array = []
        for i, o in index:
            id_ref_array.append(i)
            offset_array.append(o.offset)
        id_ref_array.append(b'')
        offset_array.append(last)

        id_ref_array_enc = bytearray(b'\x00'.join(id_ref_array))
        self.h5_file.create_dataset(
            offset_index_key, data=np.array(offset_array))
        self.h5_file.create_dataset(
            offset_index_id_ref, data=id_ref_array_enc)

    def _prepare_metadata_index(self, index, index_name, last=None, dtype=np.float32):
        n = len(index)
        value = np.empty(n + 1, dtype=dtype)
        value[:n] = index
        value[n] = dtype(last) if last is not None else dtype()
        self.h5_file.create_dataset(index_name, data=value)



from ..mzml.components import BinaryDataArray, Binary, NullMap
from ..xml import _element


EXTERNAL_DATASET_PARAM = "external HDF5 dataset"
# EXTERNAL_DATASET_PARAM = "external dataset"


class ExternalBinaryDataArray(BinaryDataArray):
    def __init__(self, external_dataset_name, data_processing_reference=None,
                 offset=None, array_length=None, params=None, context=NullMap, **kwargs):
        if (params is None):
            params = []
        self.external_dataset_name = external_dataset_name
        self.array_length = array_length
        self.offset = offset
        self.data_processing_reference = data_processing_reference
        if data_processing_reference:
            self._data_processing_reference = context[
                'DataProcessing'][data_processing_reference]
        else:
            self._data_processing_reference = None
        self.params = self.prepare_params(params, **kwargs)
        self.element = _element(
            'binaryDataArray',
            encodedLength=0,
            dataProcessingRef=self._data_processing_reference)
        self.context = context
        self._array_type = None
        self._prepare_external_refs()
        self.binary = Binary(b"", context=self.context)

    def _prepare_external_refs(self):
        self.add_param({
            "name": EXTERNAL_DATASET_PARAM,
            "value": self.external_dataset_name
        }).add_param({
            "name": "external array length",
            "value": self.array_length,
        }).add_param({
            "name": "external offset",
            "value": self.offset
        })
        return self




from io import BytesIO
from psims.transform import mzid, utils

from psims.test.test_data import datafile


def test_mzid_pipe():
    buff = BytesIO()
    path = datafile("xiFDR-CrossLinkExample_single_run.mzid")
    st = mzid.MzIdentMLTranslater(path, buff)
    st.write()

    buff.seek(0)
    test_reader = mzid.MzIdentMLParser(buff)
    ref_reader = st.reader
    ref_reader.reset()

    for tag in ["Peptide", "SpectrumIdentificationResult", "ProteinAmbiguityGroup", "SpectrumIdentificationProtocol"]:
        ref_reader.reset()
        test_reader.reset()

        for ref, test in zip(ref_reader.iterfind(tag, retrieve_refs=False),
                             test_reader.iterfind(tag, retrieve_refs=False)):
            assert utils.differ(ref, test)


if __name__ == '__main__':
    test_mzid_pipe()



import warnings
import hashlib
import os

from functools import total_ordering

from lxml import etree
from six import add_metaclass, string_types as basestring

try:
    import urlparse
except ImportError:
    from urllib import parse as urlparse

try:
    from collections.abc import Iterable, Mapping
except ImportError:
    from collections import Iterable, Mapping

from collections import OrderedDict

from psims import compression


def ensure_iterable(obj):
    """Ensure ``obj`` is either a sequential iterable object that is not a
    string type.

        1. If ``obj`` is :const:`None` return an empty :class:`tuple`.
        2. If ``obj`` is an instance of :class:`str`, :class:`bytes`, or :class:`Mapping`,
           or not :class:`Iterable` return a list containing ``obj``
        3. Return ``obj``

    Parameters
    ----------
    obj : object
        The object to ensure iterability of

    Returns
    -------
    :class:`Sequence`
        Returns either ``obj`` or a wrapepr around ``obj``
    """
    if obj is None:
        return tuple()
    if not isinstance(obj, Iterable) or isinstance(obj, basestring) or isinstance(obj, Mapping):
        return [obj]
    return obj


def is_uri(string):
    parsed = urlparse.urlparse(string)
    # No protocol
    if not parsed.scheme:
        return False
    # Windows drive path
    elif len(parsed.scheme) == 1 and parsed.scheme.isalpha():
        return False
    return True


def make_file_uri(path):
    if os.path.isabs(path):
        return "file:///" + path
    else:
        return "file://" + path


def checksum_file(path, hash_type='sha-1'):
    """Calculate the cryptographic hash checksum of the given file
    path

    Parameters
    ----------
    path : :class:`str`
        The path to the file to checksum
    hash_type : str, optional
        The name of the hash type to use. Defaults to sha-1

    Returns
    -------
    :class:`bytes`
        The hexdigest checksum of the file specified
    """
    digestor = hashlib.new(hash_type)
    with open(path, 'rb') as fh:
        chunk_size = 2 ** 16
        chunk = fh.read(chunk_size)
        while chunk:
            digestor.update(chunk)
            chunk = fh.read(chunk_size)
    return digestor.hexdigest()


def pretty_xml(path, outpath=None, encoding=b'utf-8'):
    """Format an XML document using :func:`lxml.etree.tostring` with
    ``pretty_print=True``.

    Attempts to do the right thing when given file paths and
    seekable file objects.

    Parameters
    ----------
    path : :class:`str` or file-like
        The file to format. If file-like, it will attempt to seek to the beginning
    outpath : :class:`str` or file-like, optional
        The place to write the formatted file to. If missing it will attempt to overrwrite
        the input path
    encoding : bytes, optional
        The encoding of the XML document to write out. Defaults to UTF-8
    """
    try:
        path.seek(0)
    except AttributeError:
        pass
    tree = etree.parse(path)
    if outpath is None:
        opener = compression.get(path)
        outpath = opener(path, 'wb')
    if hasattr(outpath, 'write'):
        outstream = outpath
    else:
        opener = compression.get(path)
        outstream = opener(outpath, 'wb')
    with outstream:
        # try to ensure that the stream is at the beginning
        try:
            outstream.seek(0)
        except Exception:
            pass
        formatted = etree.tostring(tree, pretty_print=True, encoding=encoding, xml_declaration=True)
        outstream.write(formatted)


def simple_repr(self):  # pragma: no cover
    template = "{self.__class__.__name__}({d})"

    def formatvalue(v):
        if isinstance(v, float):
            return "%0.4f" % v
        else:
            return str(v)

    if not hasattr(self, "__slots__"):
        d = [
            "%s=%s" % (k, formatvalue(v)) if v is not self else "(...)" for k, v in sorted(
                self.__dict__.items(), key=lambda x: x[0])
            if (not k.startswith("_") and not callable(v)) and not (v is None)]
    else:
        d = [
            "%s=%s" % (k, formatvalue(v)) if v is not self else "(...)" for k, v in sorted(
                [(name, getattr(self, name)) for name in self.__slots__], key=lambda x: x[0])
            if (not k.startswith("_") and not callable(v)) and not (v is None)]

    return template.format(self=self, d=', '.join(d))


@total_ordering
class SimpleVersion(object):
    """A simple representation of a versioning system with a three part verison
    of the form "major"."minor"."patch".

    Used to represent different schema versions and to make schema version comparison
    simpler.

    Attributes
    ----------
    major : int
        Major version
    minor : int
        Minor version
    patch : int
        Patch version
    """

    def __init__(self, major=0, minor=0, patch=0):
        self.major = major
        self.minor = minor
        self.patch = patch

    def __iter__(self):
        yield self.major
        yield self.minor
        yield self.patch

    def __hash__(self):
        return hash(tuple(self))

    def __getitem__(self, i):
        if i == 0:
            return self.major
        elif i == 1:
            return self.minor
        elif i == 2:
            return self.patch
        raise IndexError(i)

    def __len__(self):
        return 3

    def __repr__(self):
        t = "{self.__class__.__name__}({self.major}, {self.minor}, {self.patch})"
        return t.format(self=self)

    def __str__(self):
        return '.'.join(map(str, self))

    def __eq__(self, other):
        return tuple(self) == tuple(other)

    def __lt__(self, other):
        return tuple(self) < tuple(other)

    @classmethod
    def parse(cls, text):
        parts = text.split(".")
        out = []
        for a in parts:
            try:
                v = int(a)
            except (TypeError, ValueError):
                v = a
            out.append(v)
        if len(out) > 3:
            raise ValueError("Not a SimpleVersion")
        return cls(*out)


class KeyToAttrProxy(object):
    """Proxies requests for keys through :meth:`__getitem__` into
    calls to :meth:`__getattr__` and mock most methods of the :class:`Mapping`
    interface for objects with a :attr:`__dict__`

    Attributes
    ----------
    source : object
        The object to proxy
    """

    def __init__(self, source):
        self.source = source

    def __getitem__(self, key):
        try:
            return getattr(self.source, key)
        except AttributeError:
            raise KeyError(key)

    def keys(self):
        return list(filter(lambda x: not x.startswith("_"), self.source.__dict__.keys()))

    def values(self):
        return [kv[1] for kv in filter(lambda x: not x[0].startswith("_"), self.source.__dict__.items())]

    def items(self):
        return [kv for kv in filter(lambda x: not x[0].startswith("_"), self.source.__dict__.items())]

    def __contains__(self, k):
        return k in self.source.__dict__

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return len(self.keys())


class StateSpaceBase(object):
    def validate(self, start, end):
        raise NotImplementedError()

    def next_states(self, current):
        raise NotImplementedError()


class StateTable(StateSpaceBase):
    """A simple state transition table that maps a current
    state to a list of valid new states

    Attributes
    ----------
    table : OrderedDict
        The mapping from current state to next states
    """

    def __init__(self, table):
        self.table = OrderedDict(table)

    def validate(self, start, end):
        options = self.table[start]
        return end in options

    def next_states(self, current):
        try:
            return list(self.table[current])
        except KeyError:
            return []

    def __getitem__(self, i):
        return list(self.table.items())[i]

    def __len__(self):
        return len(self.states)


class StateTransitionWarning(Warning):
    pass


class StateMachineBase(object):
    """A base class implementing the logic for
    handling state transitions, state checking,
    and error reporting when an invalid transition
    or failed check occurs

    Attributes
    ----------
    current_state : object
        The current state
    enabled : bool
        Whether or not to warn about invalid actions
    """

    def __init__(self, current_state):
        self._current_state = None
        self._previous_state = None
        self.current_state = current_state
        self.enabled = True

    def transition(self, state):
        """Move from the current state to the specified state.

        If the transition was invalid and warnings are enabled,
        this will call :meth:`transition_error`

        Parameters
        ----------
        state : object
            The state to transition to

        Returns
        -------
        bool
            Whether or not the transition was valid
        """
        is_valid = self.states.validate(self.current_state, state)
        self.current_state = state
        if not is_valid and self.enabled:
            self.transition_error()
        return is_valid

    @property
    def current_state(self):
        return self._current_state

    @current_state.setter
    def current_state(self, value):
        self._previous_state = self.current_state
        self._current_state = value

    @property
    def previous_state(self):
        return self._previous_state

    def expects_state(self, state):
        """Check whether the current state matches the expected
        state

        If the expectation was invalid and warnings are enabled,
        this will call :meth:`expects_error`


        Parameters
        ----------
        state : object
            The expected state

        Returns
        -------
        bool
            Whether or not the states matched.
        """
        is_valid = self.current_state == state
        if not is_valid and self.enabled:
            self.expects_error(state)
        return is_valid

    def expects_error(self, state):
        """Warns that the expectation of `state` was invalid

        Parameters
        ----------
        state : object
            The expected state
        """
        warnings.warn(
            ("Action expected {state!r} but current state is {self.current_state!r}").format(
                self=self, state=state), StateTransitionWarning)

    def transition_error(self):
        """Warn that the last transition was invalid, and list
        the set of valid transitions.
        """
        next_states = self.states.next_states(self.previous_state)
        warnings.warn(
            ("Transition from {self.previous_state!r} to {self.current_state!r} is"
             " not valid. Expected one of {next_states!r}").format(self=self, next_states=next_states),
            StateTransitionWarning)


class TableStateMachine(StateMachineBase):
    """An implementation of :class:`StateMachineBase` which
    uses a :class:`StateTable` to represent the set of transtions
    allowed.

    Attributes
    ----------
    states : :class:`StateTable`
        The valid transitions
    """

    def __init__(self, state_table, current_state=None):
        self.states = StateTable(state_table)
        super(TableStateMachine, self).__init__(current_state)
        if current_state is None:
            self.current_state = self.states[0][0]



import os
import warnings
import operator
import re

try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping
from datetime import datetime
from numbers import Number as NumberBase
from itertools import chain

from lxml import etree

from psims.controlled_vocabulary import UNIMODEntity
from ..utils import SimpleVersion, checksum_file
from ..xml import (
    _element, element, TagBase,
    CVParam, UserParam, CV,
    ProvidedCV, sanitize_id)
from ..document import (
    ComponentBase as _ComponentBase, NullMap,
    XMLBindingDispatcherBase, ParameterContainer)

from .utils import ensure_iterable

AUTO = object()


try:
    FileNotFoundError
except Exception as e:
    FileNotFoundError = OSError


class MzIdentML(TagBase):
    v1_1_0_type_attrs = {
        'xmlns': 'http://psidev.info/psi/pi/mzIdentML/1.1',
        'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
        'xsi:schemaLocation': 'http://psidev.info/psi/pi/mzIdentML/1.1 ../../schema/mzIdentML1.1.0.xsd'
    }

    v1_1_1_type_attrs = {
        'xmlns': 'http://psidev.info/psi/pi/mzIdentML/1.1.1',
        'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
        'xsi:schemaLocation': 'http://psidev.info/psi/pi/mzIdentML/1.1.1 ../../schema/mzIdentML1.1.1.xsd'
    }

    v1_2_0_type_attrs = {
        'xmlns': 'http://psidev.info/psi/pi/mzIdentML/1.2',
        'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
        'xsi:schemaLocation': 'http://psidev.info/psi/pi/mzIdentML/1.2 ../../schema/mzIdentML1.2.0.xsd'
    }

    attr_version_map = {
        "1.1.0": v1_1_0_type_attrs,
        "1.1.1": v1_1_1_type_attrs,
        "1.2.0": v1_2_0_type_attrs,
    }

    def __init__(self, **attrs):
        attrs.setdefault('creationDate', datetime.utcnow().isoformat())
        attrs.setdefault('id', 0)
        attrs.setdefault('version', '1.1.0')
        version = attrs['version']
        self.type_attrs = self.attr_version_map[version]
        super(MzIdentML, self).__init__('MzIdentML', **attrs)


_COMPONENT_NAMESPACE = 'mzid'


class ComponentDispatcher(XMLBindingDispatcherBase):

    def __init__(self, *args, **kwargs):
        super(ComponentDispatcher, self).__init__(
            component_namespace=_COMPONENT_NAMESPACE, *args, **kwargs)

    def _update_component_namespace(self, tp=None):
        ns = super(ComponentDispatcher, self)._update_component_namespace(tp)
        ns['xmlns'] = self.xmlns
        return ns

    def _post_constructor(self, component):
        super(ComponentDispatcher, self)._post_constructor(component)


_xmlns = 'http://psidev.info/psi/pi/mzIdentML/1.2'


class ComponentBase(_ComponentBase):
    component_namespace = _COMPONENT_NAMESPACE
    xmlns = _xmlns

    @property
    def schema_version(self):
        return SimpleVersion.parse(self.xmlns.split("/")[-1])


default_cv_list = [
    CV(id='PSI-MS', uri='https://raw.githubusercontent.com/HUPO-PSI/psi-ms-CV/master/psi-ms.obo', full_name='PSI-MS'),
    CV(id='UO', uri='http://ontologies.berkeleybop.org/uo.obo', full_name='UNIT-ONTOLOGY'),
    ProvidedCV(id='UNIMOD', uri='http://www.unimod.org/obo/unimod.obo', full_name='UNIMOD',
               converter=UNIMODEntity.converter),
    CV(id='XLMOD', uri='https://raw.githubusercontent.com/HUPO-PSI/mzIdentML/master/cv/XLMOD.obo', full_name='XLMOD')
]

common_units = {
    'parts per million': 'UO:0000169',
    'dalton': 'UO:0000221'
}


class GenericCollection(ComponentBase):
    requires_id = False

    def __init__(self, tag_name, members, context=NullMap):
        self.members = members
        self.tag_name = tag_name
        self.element = _element(tag_name, xmlns=self.xmlns)

    def write_content(self, xml_file):
        for member in self.members:
            member.write(xml_file)


class IDGenericCollection(GenericCollection):
    requires_id = True

    def __init__(self, tag_name, members, id, context=NullMap):
        self.members = members
        self.tag_name = tag_name
        self.element = _element(tag_name, xmlns=self.xmlns, id=id)
        context[tag_name][id] = self.element.id

    def write_content(self, xml_file):
        for member in self.members:
            member.write(xml_file)

# ---- Input File Description ---------


class _FileDescriptionBase(object):

    @property
    def path(self):
        return self.location

    def is_local(self):
        path = self.path
        if path.startswith("file:///"):
            path = path.replace("file:///", '', 1)
        return os.path.exists(path)

    def checksum(self, digest='sha-1'):
        if not self.is_local():
            raise FileNotFoundError("Can't checksum %r, file not found")
        path = self.path
        if path.startswith("file:///"):
            path = path.replace("file:///", '', 1)
        checksum = checksum_file(path, digest)
        return self.context.param(name=digest, value=checksum)


class SourceFile(ComponentBase, _FileDescriptionBase):
    requires_id = True

    def __init__(self, location, file_format=None, id=None,
                 external_format=None, params=None, name=None, context=NullMap, **kwargs):
        self.params = self.prepare_params(params, **kwargs)
        self.external_format = external_format
        self.file_format = file_format
        self.element = _element('SourceFile', location=location, id=id, name=name)
        self.context = context
        context['SourceFile'][id] = self.element.id

    def write_content(self, xml_file):
        if self.file_format is not None:
            with element(xml_file, 'FileFormat'):
                self.write_params(xml_file, self.prepare_params(self.file_format))
        if (self.external_format is not None):
            with element(xml_file, 'ExternalFormatDocumentation'):
                xml_file.write(str(self.external_format))
        self.write_params(xml_file)


class SearchDatabase(ComponentBase, _FileDescriptionBase):
    requires_id = True

    def __init__(self, name, file_format=None, location=None, id=None, external_format=None,
                 num_database_sequences=None, num_residues=None, release_date=None, version=None,
                 params=None, context=NullMap, **kwargs):
        self.external_format = external_format
        self.params = self.prepare_params(params, **kwargs)
        self.location = location
        self.file_format = file_format
        self.element = _element(
            'SearchDatabase', location=location, name=name, id=id)
        if num_database_sequences is not None:
            self.element.attrs['numDatabaseSequences'] = int(num_database_sequences)
        if num_residues is not None:
            self.element.attrs['numResidues'] = int(num_residues)
        if release_date is not None:
            self.element.attrs['releaseDate'] = release_date
        if version is not None:
            self.element.attrs['version'] = version
        self.context = context
        self.context['SearchDatabase'][id] = self.element.id

    def write_content(self, xml_file):
        if self.file_format is not None:
            with element(xml_file, 'FileFormat'):
                self.write_params(xml_file, self.prepare_params(self.file_format))
        with element(xml_file, 'DatabaseName'):
            UserParam(name=self.name).write(xml_file)
        if (self.external_format is not None):
            with element(xml_file, 'ExternalFormatDocumentation'):
                xml_file.write(str(self.external_format))
        self.write_params(xml_file)


class SpectraData(ComponentBase, _FileDescriptionBase):
    requires_id = True

    def __init__(self, location, file_format=None, spectrum_id_format=None,
                 id=None, external_format=None, params=None, name=None, context=NullMap, **kwargs):
        self.params = self.prepare_params(params, **kwargs)
        self.external_format = external_format
        self.file_format = file_format
        self.spectrum_id_format = spectrum_id_format
        self.element = _element('SpectraData', id=id, location=location, name=name)
        context['SpectraData'][id] = self.element.id
        self.context = context

    def write_content(self, xml_file):
        if self.file_format is not None:
            with element(xml_file, 'FileFormat'):
                self.write_params(xml_file, self.prepare_params(self.file_format))
        with element(xml_file, 'SpectrumIDFormat'):
            self.context.param(self.spectrum_id_format)(xml_file)
        if (self.external_format is not None):
            with element(xml_file, 'ExternalFormatDocumentation'):
                xml_file.write(str(self.external_format))
        self.write_params(xml_file)


class Inputs(GenericCollection, ):

    def __init__(self, source_files=tuple(), search_databases=tuple(),
                 spectra_data=tuple(), context=NullMap):
        items = list()
        items.extend(source_files)
        items.extend(search_databases)
        items.extend(spectra_data)
        super(Inputs, self).__init__('Inputs', items, context=context)

# ---- Protein Database Description ---


class SequenceCollection(GenericCollection):

    def __init__(self, db_sequences, peptides,
                 peptide_evidence, context=NullMap):
        super(SequenceCollection, self).__init__('SequenceCollection',
                                                 chain.from_iterable([db_sequences, peptides, peptide_evidence]))


class DBSequence(ComponentBase):
    requires_id = True

    def __init__(self, accession, sequence=None, id=None,
                 search_database_id=1, params=None, name=None, context=NullMap, **kwargs):
        self.sequence = sequence
        self.search_database_ref = context[
            'SearchDatabase'][search_database_id]
        self.element = _element('DBSequence', accession=accession,
                                id=id, searchDatabase_ref=self.search_database_ref, name=name)
        if (sequence is not None):
            self.element.attrs['length'] = len(sequence)
            kwargs.pop("length", None)
        elif 'length' in kwargs:
            self.element.attrs['length'] = kwargs.pop('length')
        params = self.prepare_params(params, **kwargs)
        self.params = params
        context['DBSequence'][id] = self.element.id
        self.context = context

    def write_content(self, xml_file):
        if (self.sequence is not None):
            with element(xml_file, 'Seq'):
                xml_file.write(self.sequence)
        self.write_params(xml_file)


class Peptide(ComponentBase):
    requires_id = True

    def __init__(self, peptide_sequence, id, modifications=None, substitutions=None,
                 params=None, name=None, context=NullMap, **kwargs):
        params = self.prepare_params(params, **kwargs)
        self.context = context
        self.peptide_sequence = peptide_sequence
        self.modifications = [Modification.ensure(mod, context=context) for mod in ensure_iterable(modifications)]
        self.substitutions = [SubstitutionModification.ensure(sub, context=context)
                              for sub in ensure_iterable(substitutions)]
        self.params = params
        self.element = _element('Peptide', id=id, name=name)
        context['Peptide'][id] = self.element.id
        self.context = context

    def write_content(self, xml_file):
        with element(xml_file, 'PeptideSequence'):
            xml_file.write(self.peptide_sequence)
        for mod in self.modifications:
            mod.write(xml_file)
        for subst in self.substitutions:
            subst.write(xml_file)
        self.write_params(xml_file)


class ModificationDescriptionBase(object):
    UNKNOWN_MODIFICATION_ACCESSION = 'MS:1001460'

    def _resolve_name_accession(self, name):
        try:
            try:
                (mod, cv) = self.context.term(name, include_source=True)
            except TypeError:
                p = self.context.param(name)
                (mod, cv) = self.context.term(p.accession, include_source=True)
            self.name = mod['name']
            self.accession = mod['id']
            self.known = True
            self.reference = cv.id
        except KeyError:
            self.name = name
            self._make_unknown()

    def _make_unknown(self):
        self.known = False
        self.accession = self.UNKNOWN_MODIFICATION_ACCESSION
        cv = self.context.get_vocabulary('PSI-MS')
        if (cv is None):
            self.reference = 'PSI-MS'
        else:
            self.reference = cv.id

    def _format_identity(self, xml_file):
        if self.known:
            self.context.param(
                name=self.name, accession=self.accession, ref=self.reference, value=self.value)(xml_file)
        else:
            self.context.param(name='unknown modification', accession=self.accession,
                               value=self.name, ref=self.reference)(xml_file)


class Modification(ComponentBase, ModificationDescriptionBase):
    requires_id = False

    def __init__(self, monoisotopic_mass_delta=None, location=None, name=None, accession=None, params=None,
                 residues=None, value=None, avg_mass_delta=None, context=NullMap, **kwargs):
        self.context = context
        self.accession = accession
        self.name = name
        self.value = value or ""
        self.params = self.prepare_params(params, **kwargs)
        self.known = True
        if accession == self.UNKNOWN_MODIFICATION_ACCESSION and name is not None:
            self.name = name
            self._make_unknown()
        elif ((accession is None) and (name is not None)):
            self._resolve_name_accession(name)
        elif (accession is not None):
            self._resolve_name_accession(accession)
        else:
            warnings.warn(('Unknown modification saved: %s' %
                           monoisotopic_mass_delta))
            self._make_unknown()
        self.residues = residues
        self.element = _element(
            'Modification', monoisotopicMassDelta=monoisotopic_mass_delta, location=location,
            avgMassDelta=avg_mass_delta)
        if self.residues is not None:
            self.element.attrs['residues'] = ''.join(map(str, self.residues))

    def write_content(self, xml_file):
        if (self.accession is not None):
            self._format_identity(xml_file)
        self.write_params(xml_file)


class SubstitutionModification(ComponentBase):
    requires_id = False

    def __init__(self, original_residue, replacement_residue,
                 monoisotopic_mass_delta=None, location=None, context=NullMap):
        self.original_residue = original_residue
        self.replacement_residue = replacement_residue
        self.monoisotopic_mass_delta = monoisotopic_mass_delta
        self.location = location
        self.element = _element('SubstitutionModification', originalResidue=self.original_residue,
                                replacementResidue=self.replacement_residue)
        if (self.monoisotopic_mass_delta is not None):
            self.element.attrs[
                'monoisotopicMassDelta'] = self.monoisotopic_mass_delta
        if (self.location is not None):
            self.element.attrs['location'] = self.location
        self.context = context

    def write_content(self, xml_file):
        pass


class PeptideEvidence(ComponentBase):
    requires_id = True

    def __init__(self, peptide_id, db_sequence_id, id, start_position, end_position, is_decoy=False,
                 pre=None, post=None, params=None, frame=None, translation_table_id=None, context=NullMap, **kwargs):
        params = self.prepare_params(params, **kwargs)
        self.params = params
        self.context = context
        self.peptide_id = peptide_id
        self.db_sequence_id = db_sequence_id
        self.translation_table_id = translation_table_id
        self.frame = frame
        self.element = _element(
            'PeptideEvidence', isDecoy=is_decoy, start=start_position, end=end_position, peptide_ref=context[
                'Peptide'][peptide_id], dBSequence_ref=context['DBSequence'][db_sequence_id], pre=pre, post=post, id=id,
            frame=frame)
        if (self.translation_table_id is not None):
            self.element.attrs['translationTable_ref'] = self.context[
                'TranslationTable'][self.translation_table_id]
        self.context['PeptideEvidence'][id] = self.element.id

    def write_content(self, xml_file):
        self.write_params(xml_file)


# ---- Identification Results ---------

class SpectrumIdentificationResult(ComponentBase):
    requires_id = True

    def __init__(self, spectra_data_id, spectrum_id, id=None,
                 identifications=None, params=None, name=None, context=NullMap, **kwargs):
        if (identifications is None):
            identifications = []
        self.params = self.prepare_params(params, **kwargs)
        self.identifications = identifications
        self.element = _element('SpectrumIdentificationResult', spectraData_ref=context[
                                'SpectraData'][spectra_data_id], spectrumID=spectrum_id, id=id,
                                name=name)
        self.context = context
        self.context['SpectrumIdentificationResult'][id] = self.element.id

    def write_content(self, xml_file):
        for item in self.identifications:
            item.write(xml_file)
        self.after(self.write_params)


class IonType(ComponentBase):
    requires_id = False

    def __init__(self, series, indices, charge_state,
                 measures=None, params=None, context=NullMap, **kwargs):
        if (measures is None):
            measures = dict()
        self.context = context
        self.series = series
        self.measures = measures
        self.element = _element('IonType', charge=charge_state, index=' '.join(
            [str(int(i)) for i in ensure_iterable(indices)]))
        self.params = self.prepare_params(params, **kwargs)

    def write_content(self, xml_file):
        for (measure, values) in self.measures.items():
            el = _element('FragmentArray', measure_ref=self.context['Measure'][
                          measure], values=' '.join([str(float(i)) for i in values]))
            el.write(xml_file)
        self.context.param(self.series)(xml_file)
        self.write_params(xml_file)

    all_ion_types = [
        'frag: a ion', 'frag: a ion - H2O', 'frag: a ion - NH3',
        'frag: b ion', 'frag: b ion - H2O', 'frag: b ion - NH3',
        'frag: c ion', 'frag: c ion - H2O', 'frag: c ion - NH3',
        'frag: d ion',
        'frag: immonium ion',
        'frag: internal ya ion',
        'frag: internal yb ion',
        'frag: precursor ion', 'frag: precursor ion - H2O', 'frag: precursor ion - NH3',
        'frag: v ion',
        'frag: w ion',
        'frag: x ion', 'frag: x ion - H2O', 'frag: x ion - NH3',
        'frag: y ion', 'frag: y ion - H2O', 'frag: y ion - NH3',
        'frag: z ion', 'frag: z ion - H2O', 'frag: z ion - NH3',
        'frag: z+1 ion', 'frag: z+2 ion'
    ]

    _ion_pattern = re.compile(
        r"""(?P<prefix>frag:\s?)?
            (?P<key>[abcxyzdvw]|internal [yab]{2}|precursor|immonium|z\+[12])
            (?P<ion>\s?ion\s?)?
            (?=\s?-\s?(?P<loss>NH3|H2O))?""", re.VERBOSE)

    @classmethod
    def guess_ion_type(cls, string):
        match = cls._ion_pattern.search(string)
        if match:
            groups = match.groupdict()
            loss = groups.get('loss')
            if (loss is not None):
                loss = (' - ' + loss)
            else:
                loss = ''
            return 'frag: {key} ion{loss}'.format(key=groups['key'], loss=loss)
        else:
            raise KeyError(('No mapping found for %s' % string))


class SpectrumIdentificationItem(ComponentBase):
    requires_id = True

    def __init__(self, experimental_mass_to_charge, charge_state, peptide_id, peptide_evidence_ids, id, score=None,
                 ion_types=None, params=None, pass_threshold=True, rank=1, calculated_mass_to_charge=None,
                 calculated_pi=None, name=None, mass_table_id=None, sample_id=None, context=NullMap, **kwargs):
        params = self.prepare_params(params, **kwargs)
        self.peptide_evidence_refs = [context['PeptideEvidence'][
            peptide_evidence_id] for peptide_evidence_id in ensure_iterable(peptide_evidence_ids)]
        self.params = params
        self.score = score
        self.rank = rank
        self.element = _element(
            'SpectrumIdentificationItem', chargeState=charge_state,
            experimentalMassToCharge=experimental_mass_to_charge,
            id=id, passThreshold=pass_threshold, peptide_ref=context['Peptide'][peptide_id], rank=self.rank,
            name=name)
        self.element.attrs[
            'calculatedMassToCharge'] = calculated_mass_to_charge
        self.element.attrs['calculatedPI'] = calculated_pi
        if (sample_id is not None):
            self.element.attrs['sample_ref'] = context['Sample'][sample_id]
        if (mass_table_id is not None):
            self.element.attrs['massTable_ref'] = context[
                'MassTable'][sample_id]
        context['SpectrumIdentificationItem'][id] = self.element.id
        self.context = context
        self.ion_types = self.prepare_ion_types(ion_types)

    def prepare_ion_types(self, ion_types):
        mappings = []
        if (ion_types is not None):
            if isinstance(ion_types, (list, tuple)):
                try:
                    if isinstance(ion_types[0], IonType):
                        return ion_types
                except IndexError:
                    return mappings
            elif isinstance(ion_types, Mapping):
                ion_types = list(ion_types.items())
            for (series, measures) in ion_types:
                measures_ = {}
                for (measure, values) in measures.items():
                    if (measure in ('indices', 'charge_state')):
                        continue
                    measures_[measure] = values
                it = IonType(series=IonType.guess_ion_type(series), indices=measures[
                             'indices'], charge_state=measures.get('charge_state', 1), measures=measures_,
                             context=self.context)
                mappings.append(it)
        return mappings

    def write_content(self, xml_file):
        for peptide_evidence_ref in self.peptide_evidence_refs:
            _element('PeptideEvidenceRef',
                     peptideEvidence_ref=peptide_evidence_ref).write(xml_file)
        ion_types = ensure_iterable(self.ion_types)
        if ((self.ion_types is not None) and (len(ion_types) > 0)):
            with element(xml_file, 'Fragmentation'):
                for ion_type in ion_types:
                    ion_type.write(xml_file)
        if isinstance(self.score, CVParam):
            self.score.write(xml_file)
        elif isinstance(self.score, dict):
            self.context.param(self.score)(xml_file)
        elif self.score is None:
            pass
        else:
            self.context.param(name='score', value=self.score)(xml_file)
        self.write_params(xml_file)


class Measure(ComponentBase):
    requires_id = True

    def __init__(self, name=None, id=None, param=None,
                 getter=None, context=NullMap):
        if isinstance(name, CVParam):
            param = name
            name = param.name
        elif ((param is None) and (name is not None)):
            param = context.param(name)
        self.name = name
        self.param = param
        if ((name is not None) and (id is None)):
            id = name
        self.element = _element('Measure', id=sanitize_id(id), name=name)
        context['Measure'][id] = self.element.id
        self.context = context
        if (getter is None):
            getter = operator.itemgetter(name)
        self.getter = getter

    def write_content(self, xml_file):
        self.context.param(self.param)(xml_file)

    common_measures = [({'name': 'product ion %s' % s, 'unit_name': u})
                       for s, u in (('m/z', 'm/z'), ('intensity', 'number of detector counts'), (
                                    'm/z error', 'm/z'))]


class FragmentationTable(ComponentBase):
    requires_id = False

    def __init__(self, measures=None, context=NullMap):
        if (measures is None):
            measures = Measure.common_measures
        self.context = context
        self.element = _element('FragmentationTable')
        self.measures = []
        for measure in measures:
            if isinstance(measure, Measure):
                self.measures.append(measure)
            else:
                measure = Measure(context.param(measure), context=context)
                self.measures.append(measure)

    def write_content(self, xml_file):
        for measure in self.measures:
            measure.write(xml_file)


class SpectrumIdentificationList(ComponentBase):
    requires_id = True

    def __init__(self, identification_results, id,
                 fragmentation_table=None, params=None, context=NullMap, name=None, **kwargs):
        self.identification_results = identification_results
        self.fragmentation_table = fragmentation_table
        self.element = _element(
            'SpectrumIdentificationList', xmlns=self.xmlns, id=id, name=name)
        self.context = context
        self.context['SpectrumIdentificationList'][id] = self.element.id
        self.params = self.prepare_params(params, **kwargs)

    def write_content(self, xml_file):
        self.write_params(xml_file)
        if (self.fragmentation_table is not None):
            self.fragmentation_table.write(xml_file)
        for identification_result in self.identification_results:
            identification_result.write(xml_file)


class AnalysisData(GenericCollection):

    def __init__(self, identification_lists=tuple(),
                 protein_detection_lists=tuple(), context=NullMap):
        items = list()
        items.extend(identification_lists)
        items.extend(protein_detection_lists)
        super(AnalysisData, self).__init__('AnalysisData', items, context)


class ProteinDetectionList(ComponentBase):
    requires_id = True

    def __init__(self, ambiguity_groups=None, count=AUTO,
                 id=1, params=None, context=NullMap, **kwargs):
        self.ambiguity_groups = ensure_iterable(ambiguity_groups)
        self.element = _element('ProteinDetectionList', id=id)
        self.count = count
        self.params = self.prepare_params(params, **kwargs)
        self.context = context
        self.context['ProteinDetectionList'][id] = self.element.id

    def _count_protein_groups(self):
        count = 0
        for pg in self.ambiguity_groups:
            if pg.pass_threshold:
                count += 1
        return count

    def write_content(self, xml_file):
        for ambiguity_group in self.ambiguity_groups:
            ambiguity_group.write(xml_file)
        if (self.count is AUTO):
            count = self._count_protein_groups()
            if (count > 0):
                self.context.param(
                    'count of identified proteins', count)(xml_file)
        elif (self.count is not None):
            self.context.param(
                'count of identified proteins', self.count)(xml_file)
        self.write_params(xml_file)


class PeptideHypothesis(ComponentBase):
    requires_id = False

    def __init__(self, peptide_evidence_id, spectrum_identification_ids,
                 params=None, context=NullMap, **kwargs):
        self.context = context
        params = self.prepare_params(params, **kwargs)
        self.params = params
        self.peptide_evidence_id = peptide_evidence_id
        self._peptide_evidence_id = self.context['PeptideEvidence'][peptide_evidence_id]
        self.spectrum_identification_ids = ensure_iterable(spectrum_identification_ids)
        self._spectrum_identification_ids = [
            self.context['SpectrumIdentificationItem'][spectrum_identification_id]
            for spectrum_identification_id in self.spectrum_identification_ids
        ]
        self.element = _element('PeptideHypothesis', peptideEvidence_ref=self._peptide_evidence_id)

    def write_content(self, xml_file):
        for spectrum_identification_id in self._spectrum_identification_ids:
            el = _element(
                'SpectrumIdentificationItemRef',
                spectrumIdentificationItem_ref=spectrum_identification_id)
            el.write(xml_file)
        self.after(self.write_params)


class ProteinDetectionHypothesis(ComponentBase):
    requires_id = True

    def __init__(self, id, db_sequence_id, peptide_hypotheses, pass_threshold=True,
                 leading=None, params=None, name=None, context=NullMap, **kwargs):
        params = self.prepare_params(params, **kwargs)
        self.db_sequence_id = db_sequence_id
        self.pass_threshold = pass_threshold
        self.leading = leading
        self.params = params
        self.context = context
        self.peptide_hypotheses = self._coerce_peptide_hypotheses(
            peptide_hypotheses)
        self.element = _element('ProteinDetectionHypothesis', id=id, dBSequence_ref=context[
                                'DBSequence'][db_sequence_id], passThreshold=pass_threshold)
        if (name is not None):
            self.element.attrs['name'] = name
        self.context['ProteinDetectionHypothesis'][id] = self.element.id
        if self.leading is not None:
            self._check_group_status()

    def _check_group_status(self):
        candidates = []
        has_leading = False
        has_non_leading = False
        for param in self.params:
            param = self.context.param(param)
            if param.accession is None:
                continue
            try:
                term = self.context.term(param.accession)
            except KeyError:
                continue
            if term.is_of_type('MS:1001101'):
                candidates.append(term)
            if term.id == 'MS:1002401':
                has_leading = True
            elif term.id == 'MS:1002402':
                has_non_leading = True
        if self.leading and not has_leading:
            self.add_param("MS:1002401")
        elif not self.leading and not has_non_leading:
            self.add_param("MS:1002402")

    def _coerce_peptide_hypotheses(self, peptides):
        out = []
        for peptide in ensure_iterable(peptides):
            if isinstance(peptide, Mapping):
                peptide = PeptideHypothesis(context=self.context, **peptide)
            out.append(peptide)
        return out

    def write_content(self, xml_file):
        for peptide_hypothesis in self.peptide_hypotheses:
            peptide_hypothesis.write(xml_file)

        self.after(self.write_params)


class ProteinAmbiguityGroup(ComponentBase):
    requires_id = True

    def __init__(self, id, protein_detection_hypotheses,
                 pass_threshold=True, params=None, context=NullMap, **kwargs):
        params = self.prepare_params(params, **kwargs)
        self.protein_detection_hypotheses = protein_detection_hypotheses
        self.params = params
        self.pass_threshold = pass_threshold
        self.context = context
        self.element = _element('ProteinAmbiguityGroup', id=id)
        self.context['ProteinAmbiguityGroup'][id] = self.element.id
        self._check_threshold()

    def _check_threshold(self):
        if self.pass_threshold is not None:
            if not (self.has_param("protein group passes threshold")):
                self.add_param({"name": 'protein group passes threshold', "value": self.pass_threshold})

    def write_content(self, xml_file):
        for protein in self.protein_detection_hypotheses:
            protein.write(xml_file)
        self.after(self.write_params)


class DataCollection(GenericCollection):

    def __init__(self, inputs, analysis_data, context=NullMap):
        super(DataCollection, self).__init__(
            'DataCollection', [inputs, analysis_data], context)


#


class MassTable(ComponentBase):
    requires_id = True

    def __init__(self, id, ms_level, residues=None, ambiguous_residues=None,
                 name=None, params=None, context=NullMap, **kwargs):
        if (residues is None):
            residues = []
        if (ambiguous_residues is None):
            ambiguous_residues = []
        self.context = context
        self.residues = [Residue.ensure(r, context=context)
                         for r in ensure_iterable(residues)]
        self.ambiguous_residues = [AmbiguousResidue.ensure(
            r, context=context) for r in ensure_iterable(ambiguous_residues)]
        self.name = name
        self.ms_level = list(
            map((lambda x: str(int(x))), ensure_iterable(ms_level)))
        self.element = _element(
            'MassTable', id=id, msLevel=' '.join(self.ms_level))
        if (self.name is not None):
            self.element.attrs['name'] = str(self.name)
        self.context['MassTable'][id] = self.element.id
        self.params = self.prepare_params(params, **kwargs)

    def write_content(self, xml_file):
        for residue in self.residues:
            residue.write(xml_file)
        for residue in self.ambiguous_residues:
            residue.write(xml_file)
        self.write_params(xml_file)


class Residue(ComponentBase):
    requires_id = False

    def __init__(self, mass, code, context=NullMap):
        self.mass = float(mass)
        self.code = str(code)
        self.element = _element('Residue', mass=self.mass, code=self.code)

    def write_content(self, xml_file):
        self.element.write(xml_file, with_id=False)


class AmbiguousResidue(ComponentBase):
    requires_id = False

    def __init__(self, code, params=None, context=NullMap, **kwargs):
        self.context = context
        self.code = str(code)
        self.params = self.prepare_params(params, **kwargs)
        self.element = _element('AmbiguousResidue', code=self.code)

    def write_content(self, xml_file):
        self.write_params(xml_file)


class Enzyme(ComponentBase):
    requires_id = True

    def __init__(self, name, missed_cleavages=1, id=None, semi_specific=False, site_regexp=None,
                 min_distance=None, n_term_gain=None, c_term_gain=None, params=None, context=NullMap, **kwargs):
        self.name = name
        if (site_regexp is None):
            try:
                p = context.param(name)
                term = context.term(p.name)
                regex_ref = term['has_regexp']
                regex_ent = context.term(regex_ref)
                regex = regex_ent['name']
                site_regexp = regex
            except Exception:
                pass
        self.site_regexp = site_regexp
        self.element = _element(
            'Enzyme', semiSpecific=semi_specific, missedCleavages=missed_cleavages, id=id)
        context['Enzyme'][id] = self.element.id
        self.context = context
        self.params = self.prepare_params(params, **kwargs)
        self.min_distance = min_distance
        if (min_distance is not None):
            self.element.attrs['minDistance'] = int(min_distance)
        self.n_term_gain = n_term_gain
        if (n_term_gain is not None):
            self.element.attrs['nTermGain'] = str(n_term_gain)
        self.c_term_gain = c_term_gain
        if (c_term_gain is not None):
            self.element.attrs['cTermGain'] = str(c_term_gain)

    def write_content(self, xml_file):
        if (self.site_regexp is not None):
            regex = _element('SiteRegexp').element()
            regex.text = etree.CDATA(self.site_regexp)
            xml_file.write(regex)
        with element(xml_file, 'EnzymeName'):
            self.context.param(self.name)(xml_file)
        self.write_params(xml_file)


class Enzymes(ComponentBase):
    requires_id = False

    def __init__(self, enzymes=None, independent=None, context=NullMap):
        self.independent = independent
        self.context = context
        self.enzymes = self._coerce_enzymes(enzymes)
        self.element = _element('Enzymes')
        if (self.independent is not None):
            self.element.attrs['independent'] = bool(self.independent)

    def _coerce_enzymes(self, enzymes):
        temp = []
        for enz in ensure_iterable(enzymes):
            if (not isinstance(enz, Enzyme)):
                enz = Enzyme(context=self.context, **enz)
            temp.append(enz)
        return temp

    def write_content(self, xml_file):
        for enzyme in self.enzymes:
            enzyme.write(xml_file)


class _Tolerance(ComponentBase):
    requires_id = False

    def __init__(self, low, high=None, unit='parts per million',
                 context=NullMap):
        self.context = context
        if isinstance(low, NumberBase):
            low = context.param(unit_name=unit, value=low, name='search tolerance minus value')
        if (high is None):
            high = context.param(unit_name=unit, value=low.value, name='search tolerance plus value')
        elif isinstance(high, NumberBase):
            high = context.param(unit_name=unit, value=high, name='search tolerance plus value')
        self.low = self.context.param(low)
        self.high = self.context.param(high)
        self.element = _element(self.tag_name)

    def write_content(self, xml_file):
        self.low.write(xml_file)
        self.high.write(xml_file)

    def __iter__(self):
        (yield self.low.value)
        (yield self.high.value)


class FragmentTolerance(_Tolerance, ):
    tag_name = 'FragmentTolerance'


class ParentTolerance(_Tolerance, ):
    tag_name = 'ParentTolerance'


class Threshold(ComponentBase):
    requires_id = False
    no_threshold = CVParam(accession='MS:1001494',
                           ref='PSI-MS', name='no threshold')

    def __init__(self, params=None, context=NullMap, **kwargs):
        self.context = context
        params = self.prepare_params(ensure_iterable(params), **kwargs)
        self.params = params
        self.element = _element("Threshold")

    def write_content(self, xml_file):
        if (not self.params):
            self.no_threshold(xml_file)
        else:
            self.write_params(xml_file)


class SpectrumIdentificationProtocol(ComponentBase):
    requires_id = True

    def __init__(self, search_type, analysis_software_id=1, id=1, additional_search_params=None,
                 modification_params=None, enzymes=None, fragment_tolerance=None, parent_tolerance=None,
                 threshold=None, filters=None, mass_tables=None, database_translation=None, name=None,
                 context=NullMap):
        self.context = context
        if (threshold is None):
            threshold = Threshold(context=context)
        elif (not isinstance(threshold, Threshold)):
            threshold = Threshold(threshold, context=context)
        if (not isinstance(enzymes, Enzymes)):
            enzymes = Enzymes(ensure_iterable(enzymes), context=context)
        if (database_translation is not None):
            database_translation = DatabaseTranslation.ensure(
                database_translation, context=self.context)
        self.parent_tolerance = self._prepare_tolerance_type(
            parent_tolerance, ParentTolerance)
        self.fragment_tolerance = self._prepare_tolerance_type(
            fragment_tolerance, FragmentTolerance)
        self.threshold = threshold
        self.enzymes = enzymes
        self.search_type = search_type
        self.modification_params = self._prepare_search_modifications(
            modification_params)
        self.filters = self._prepare_filters(filters)
        self.additional_search_params = self.prepare_params(
            additional_search_params)
        self._check_search_type()
        self._check_additional_search_params()
        self.mass_tables = [MassTable.ensure(
            table, context=context) for table in ensure_iterable(mass_tables)]
        self.database_translation = database_translation
        self.element = _element('SpectrumIdentificationProtocol', id=id, analysisSoftware_ref=context[
                                'AnalysisSoftware'][analysis_software_id], name=name)
        self.context['SpectrumIdentificationProtocol'][id] = self.element.id

    def _prepare_tolerance_type(self, tolerance, tolerance_class):
        if (not isinstance(tolerance, tolerance_class)):
            if isinstance(tolerance, NumberBase):
                tolerance = tolerance_class(tolerance)
            elif isinstance(tolerance, (tuple, list)):
                tolerance = tolerance_class(*tolerance)
            elif (tolerance is None):
                pass
            else:
                raise ValueError(('Cannot infer %r from %r' %
                                  (tolerance_class.__name__, tolerance)))
        return tolerance

    def _prepare_search_modifications(self, search_modifications):
        temp = []
        for mod in ensure_iterable(search_modifications):
            if isinstance(mod, SearchModification):
                temp.append(mod)
            else:
                temp.append(SearchModification(context=self.context, **mod))
        search_modifications = temp
        return search_modifications

    def _prepare_filters(self, filters):
        temp = []
        for filt in ensure_iterable(filters):
            if isinstance(filt, Filter):
                temp.append(filt)
            else:
                temp.append(Filter(context=self.context, **filt))
        return temp

    def _check_search_type(self):
        cv = self.context.get_vocabulary('PSI-MS')
        search_types = cv['MS:1001080'].children
        search_type_param = self.context.param(self.search_type)
        search_type = cv[search_type_param.name]
        if (search_type not in search_types):
            warnings.warn(
                'Search Type {} is not recognized.'.format(search_type.name))

    def _check_additional_search_params(self):
        if self.schema_version < SimpleVersion(1, 2):
            return
        special_processing_instructions = []
        cv = self.context.get_vocabulary('PSI-MS')
        special_processing_type_terms = cv['MS:1002489'].children
        special_processing_types = {
            t.id for t in special_processing_type_terms}
        for param in self.additional_search_params:
            try:
                param = self.context.param(param)
            except KeyError:
                continue
            accession = param.get('accession', None)
            if (accession in special_processing_types):
                special_processing_instructions.append(param)
        if (not special_processing_instructions):
            self.additional_search_params.insert(
                0, self.context.param('no special processing'))

    def write_content(self, xml_file):
        with element(xml_file, 'SearchType'):
            self.context.param(self.search_type)(xml_file)
        if self.additional_search_params:
            with element(xml_file, 'AdditionalSearchParams'):
                self.write_params(xml_file, self.additional_search_params)
        if self.modification_params:
            with element(xml_file, 'ModificationParams'):
                for mod in self.modification_params:
                    mod.write(xml_file)
        self.enzymes.write(xml_file)
        for mass_table in self.mass_tables:
            mass_table.write(xml_file)
        if (self.fragment_tolerance is not None):
            self.fragment_tolerance.write(xml_file)
        if (self.parent_tolerance is not None):
            self.parent_tolerance.write(xml_file)
        self.threshold.write(xml_file)
        if self.filters:
            with element(xml_file, 'DatabaseFilters'):
                for filt in self.filters:
                    filt.write(xml_file)
        if (self.database_translation is not None):
            self.database_translation.write(xml_file)


class Filter(ComponentBase):
    requires_id = False

    def __init__(self, filter_type, include=None,
                 exclude=None, context=NullMap):
        self.filter_type = filter_type
        self.include = include
        self.exclude = exclude
        self.context = context
        self.element = _element('Filter')

    def write_content(self, xml_file):
        with element(xml_file, 'FilterType'):
            self.context.param(self.filter_type)(xml_file)
        if self.include:
            with element(xml_file, 'Include'):
                self.write_params(xml_file, self.prepare_params(self.include))
        if self.exclude:
            with element(xml_file, 'Exclude'):
                self.write_params(xml_file, self.prepare_params(self.exclude))


class TranslationTable(ComponentBase):
    requires_id = True

    def __init__(self, id, name=None, params=None, context=NullMap, **kwargs):
        self.context = context
        self.name = name
        self.element = _element('TranslationTable', id=id, name=name)
        self.context['TranslationTable'][id] = self.element.id
        self.params = self.prepare_params(params, **kwargs)

    def write_content(self, xml_file):
        self.write_params(xml_file)


class DatabaseTranslation(ComponentBase):
    requires_id = False

    def __init__(self, translation_tables, frames=None, context=NullMap):
        self.frames = map(int, ensure_iterable(frames))
        frames = ' '.join(map(str, self.frames))
        self.element = _element('DatabaseTranslation', frames=frames)
        self.translation_tables = [TranslationTable.ensure(
            tt, context=context) for tt in ensure_iterable(translation_tables)]
        self._check_frames()

    def _check_frames(self):
        invalid_frames = []
        for f in self.frames:
            if (f not in ((-3), (-2), (-1), 1, 2, 3)):
                invalid_frames.append(f)
        if invalid_frames:
            warnings.warn(('Invalid translation frames %r' %
                           (invalid_frames,)))
        if (not self.frames):
            warnings.warn('No translation frames were specified')

    def write_content(self, xml_file):
        for tt in self.translation_tables:
            tt.write(xml_file)


class SpecificityRules(ParameterContainer):

    def __init__(self, params=None, context=NullMap, **kwargs):
        params = self.prepare_params(params, **kwargs)
        super(SpecificityRules, self).__init__(
            'SpecificityRules', params, context=context)


class SearchModification(ComponentBase, ModificationDescriptionBase):
    requires_id = False

    def __init__(self, mass_delta, fixed, residues, name=None, accession=None, value=None,
                 specificity=None, params=None, context=NullMap, **kwargs):
        self.context = context
        if value is None:
            value = ''
        specificity = ensure_iterable(specificity)
        if (specificity):
            if (not isinstance(specificity, (tuple, list))):
                specificity = [specificity]
            if (not isinstance(specificity[0], SpecificityRules)):
                specificity = [SpecificityRules.ensure(
                    s, context=context) for s in specificity]
        if (accession is not None):
            self._resolve_name_accession(accession)
        elif (name is not None):
            self._resolve_name_accession(name)
        else:
            self.name = None
            self.accession = None
            self.reference = None
            self.known = None
        self.value = value
        self.params = self.prepare_params(params, **kwargs)
        self.mass_delta = mass_delta
        self.fixed = fixed
        self.residues = ''.join(residues)
        self.specificity = specificity
        self.element = _element(
            'SearchModification', fixedMod=fixed, massDelta=mass_delta, residues=self.residues)

    def write_content(self, xml_file):
        if (self.specificity):
            for spec in self.specificity:
                spec.write(xml_file)
        if (self.name is not None):
            self._format_identity(xml_file)
        self.write_params(xml_file)


class ProteinDetectionProtocol(ComponentBase):
    requires_id = True

    def __init__(self, id=1, params=None, analysis_software_id=1,
                 threshold=None, name=None, context=NullMap, **kwargs):
        if (threshold is None):
            threshold = Threshold(context=context)
        elif (not isinstance(threshold, Threshold)):
            threshold = Threshold(threshold, context=context)
        params = self.prepare_params(params, **kwargs)
        self.context = context
        self.threshold = threshold
        self.params = params
        self.analysis_software_id = analysis_software_id
        self.element = _element('ProteinDetectionProtocol', id=id, analysisSoftware_ref=context[
                                'AnalysisSoftware'][analysis_software_id], name=name)
        self.context['ProteinDetectionProtocol'][id] = self.element.id

    def write_content(self, xml_file):
        self.threshold.write(xml_file)
        if self.params:
            with element(xml_file, 'AnalysisParams'):
                self.write_params(xml_file)


class AnalysisProtocolCollection(GenericCollection, ):

    def __init__(self, spectrum_identification_protocols=tuple(),
                 protein_detection_protocols=tuple(), context=NullMap):
        items = list()
        items.extend(spectrum_identification_protocols)
        items.extend(protein_detection_protocols)
        super(AnalysisProtocolCollection, self).__init__(self, items, context)


class SpectrumIdentification(ComponentBase):
    requires_id = True

    def __init__(self, spectra_data_ids_used=None, search_database_ids_used=None,
                 spectrum_identification_list_id=1, spectrum_identification_protocol_id=1,
                 id=1, activity_date=None, name=None, context=NullMap):
        self.spectra_data_ids_used = [context['SpectraData'][
            x] for x in (spectra_data_ids_used or [])]
        self.search_database_ids_used = [context['SearchDatabase'][
            x] for x in (search_database_ids_used or [])]
        self.element = _element(
            'SpectrumIdentification', id=id,
            spectrumIdentificationList_ref=context['SpectrumIdentificationList'][spectrum_identification_list_id],
            spectrumIdentificationProtocol_ref=context['SpectrumIdentificationProtocol'][
                spectrum_identification_protocol_id],
            name=name, activityDate=activity_date)
        self.context = context
        self.context['SpectrumIdentification'][id] = self.element.id

    def write_content(self, xml_file):
        for spectra_data_id in self.spectra_data_ids_used:
            _element('InputSpectra', spectraData_ref=spectra_data_id).write(
                xml_file)
        for search_database_id in self.search_database_ids_used:
            _element('SearchDatabaseRef',
                     searchDatabase_ref=search_database_id).write(xml_file)


class ProteinDetection(ComponentBase):
    requires_id = True

    def __init__(self, spectrum_identification_ids_used, protein_detection_list_id=1,
                 protein_detection_protocol_id=1, id=1, activity_date=None, name=None, context=NullMap):
        self.spectrum_identification_ids_used = [context['SpectrumIdentificationList'][
            x] for x in (spectrum_identification_ids_used or [])]
        self.protein_detection_list_id = protein_detection_list_id
        self.protein_detection_protocol_id = protein_detection_protocol_id
        self.context = context
        self.element = _element(
            'ProteinDetection', id=id,
            proteinDetectionProtocol_ref=context['ProteinDetectionProtocol'][
                protein_detection_protocol_id],
            proteinDetectionList_ref=context['ProteinDetectionList'][protein_detection_list_id],
            name=name, activityDate=activity_date)
        self.context['ProteinDetection'][id] = self.element.id

    def write_content(self, xml_file):
        for sid in self.spectrum_identification_ids_used:
            _element('InputSpectrumIdentifications',
                     spectrumIdentificationList_ref=sid).write(xml_file)


# ---- Auditing and Provenance --------

DEFAULT_CONTACT_ID = 'PERSON_DOC_OWNER'
DEFAULT_ORGANIZATION_ID = 'ORG_DOC_OWNER'


class CVList(ComponentBase):
    requires_id = False

    def __init__(self, cv_list=None, context=NullMap):
        if (cv_list is None):
            cv_list = default_cv_list
        self.cv_list = cv_list
        self.element = _element("cvList")

    def write_content(self, xml_file):
        for member in self.cv_list:
            tag = _element('cv', id=member.id,
                           fullName=member.full_name, uri=member.uri)
            if (member.version is not None):
                tag.attrs['version'] = member.version
            if member.options:
                tag.attrs.update(member.options)
            xml_file.write(tag.element(with_id=True))

    def __iter__(self):
        return iter(self.cv_list)


class AnalysisSoftwareList(GenericCollection):
    def __init__(self, analysis_software_list, context=NullMap):
        items = list(analysis_software_list)
        super(AnalysisSoftwareList, self).__init__('AnalysisSoftwareList', items, context)


class AnalysisSoftware(ComponentBase):
    requires_id = True

    def __init__(self, name, id=1, version=None, uri=None, contact=DEFAULT_CONTACT_ID,
                 role='software vendor', customization=None, context=NullMap, **kwargs):
        self.name = name
        self.version = version
        self.uri = uri
        self.contact = contact
        self.role = role
        self.kwargs = kwargs
        if isinstance(name, Mapping):
            attr_name = list(name)[0]
        else:
            attr_name = name
        self.element = _element(
            'AnalysisSoftware', id=id, name=attr_name, version=self.version, uri=self.uri)
        self.context = context
        self.context['AnalysisSoftware'][id] = self.element.id
        self.customization = customization

    def write_content(self, xml_file):
        if self.role:
            with element(xml_file, 'ContactRole', contact_ref=self.contact):
                with element(xml_file, 'Role'):
                    self.write_params(xml_file, self.prepare_params(self.role))
        with element(xml_file, 'SoftwareName'):
            name_param = self.context.param(name=self.name)
            if isinstance(name_param, UserParam):
                name_param = self.context.param('MS:1000799', value=self.name)
            name_param(xml_file)
        if (self.customization is not None):
            with element(xml_file, 'Customizations'):
                xml_file.write('\n'.join(ensure_iterable(self.customization)))


class Provider(ComponentBase):
    requires_id = True

    def __init__(self, id='PROVIDER', role='researcher', analysis_software_id=None,
                 contact=DEFAULT_CONTACT_ID, context=NullMap, **kwargs):
        self.id = id
        self.contact = contact
        self.role = role
        self.context = context
        self.analysis_software_id = analysis_software_id
        if analysis_software_id is not None:
            self._analysis_software_id = self.context['AnalysisSoftware'][analysis_software_id]
        self.element = _element('Provider', id=id, xmlns=self.xmlns)
        if analysis_software_id is not None:
            self.element.attrs['analysisSoftware_ref'] = self._analysis_software_id

        self.context['Provider'][id] = self.element.id

    def write_content(self, xml_file):
        with element(xml_file, 'ContactRole', contact_ref=self.contact):
            with element(xml_file, 'Role'):
                self.write_params(xml_file, self.prepare_params(self.role))


class Person(ComponentBase):
    requires_id = True

    def __init__(self, first_name=None, last_name=None, middle_initial=None, id=DEFAULT_CONTACT_ID,
                 affiliations=DEFAULT_ORGANIZATION_ID, params=None, context=NullMap, **kwargs):
        self.params = self.prepare_params(params, **kwargs)
        self.first_name = first_name
        self.middle_initial = middle_initial
        self.last_name = last_name
        self.id = id
        self.element = _element('Person', firstName=first_name,
                                lastName=last_name, midInitials=middle_initial, id=id)
        self.context = context
        self.context['Person'][id] = self.element.id
        self.affiliations = [self.context['Organization'][affiliation]
                             for affiliation in ensure_iterable(affiliations)]

    def write_content(self, xml_file):
        self.write_params(xml_file)
        if self.affiliations:
            for affiliation in self.affiliations:
                _element('Affiliation', organization_ref=affiliation).write(xml_file)


class Organization(ComponentBase):
    requires_id = True

    def __init__(self, name=None, id=DEFAULT_ORGANIZATION_ID,
                 params=None, parent=None, context=NullMap, **kwargs):
        self.name = name
        self.id = id
        self.params = self.prepare_params(params, **kwargs)
        self.element = _element('Organization', name=name, id=id)
        self.context = context
        self.context['Organization'][id] = self.element.id
        self.parent = parent
        if parent:
            self._parent = self.context['Organization'][parent]

    def write_content(self, xml_file):
        self.write_params(xml_file)
        if (self.parent is not None):
            _element('Parent', organization_ref=self._parent).write(xml_file)


class AuditCollection(ComponentBase):
    requires_id = False

    def __init__(self, persons=None, organizations=None, context=NullMap):
        self.context = context
        self.persons = Person.ensure_all(persons, context=context)
        self.organizations = Organization.ensure_all(organizations, context=context)
        self.element = _element("AuditCollection", xmlns=self.xmlns)

    def write_content(self, xml_file):
        for person in self.persons:
            person.write(xml_file)
        for organization in self.organizations:
            organization.write(xml_file)


class AnalysisSampleCollection(ComponentBase):
    requires_id = False

    def __init__(self, samples=None, context=NullMap):
        self.samples = ensure_iterable(samples)
        self.context = context
        self.element = _element('AnalysisSampleCollection')

    def write_content(self, xml_file):
        for sample in self.samples:
            sample.write(xml_file)


class Sample(ComponentBase):
    requires_id = True

    def __init__(self, id, name=None, contact=None, role=None,
                 sub_samples=None, params=None, context=NullMap, **kwargs):
        self.id = id
        self.name = name
        self.contact = contact
        self.role = ensure_iterable(role)
        self.sub_samples = ensure_iterable(sub_samples)
        self.params = self.prepare_params(params, **kwargs)
        self.context = context
        self.element = _element('Sample', id=self.id, name=self.name)
        self.context['Sample'][self.id] = self.element.id

    def write_content(self, xml_file):
        for role in self.role:
            with element(xml_file, 'ContactRole', contact_ref=self.contact):
                with element(xml_file, 'Role'):
                    self.write_params(xml_file, self.prepare_params(role))
        for sample_ref in self.sub_samples:
            _element('SubSample', sample_ref=self.context[
                     'Sample'][sample_ref]).write(xml_file)
        self.write_params(xml_file)


class BibliographicReference(ComponentBase):
    def __init__(self, id, authors=None, doi=None, editor=None, issue=None, names=None, pages=None,
                 publication=None, publisher=None, title=None, volume=None, year=None, context=NullMap):
        self.id = id
        self.element = _element("BibliographicReference", id=id)
        if authors:
            self.element.attrs['authors'] = authors
        if doi:
            self.element.attrs['doi'] = doi
        if editor:
            self.element.attrs['editor'] = editor
        if issue:
            self.element.attrs['issue'] = issue
        if names:
            self.element.attrs['names'] = names
        if pages:
            self.element.attrs['pages'] = pages
        if publication:
            self.element.attrs['publication'] = publication
        if publisher:
            self.element.attrs['publisher'] = publisher
        if title:
            self.element.attrs['title'] = title
        if volume:
            self.element.attrs['volume'] = volume
        if year:
            self.element.attrs['year'] = year
        self.context['BibliographicReference'][id] = self.element.id

    def write_content(self, xml_file):
        pass



import warnings
from collections import defaultdict, OrderedDict
from functools import partial, update_wrapper
from contextlib import contextmanager

from .controlled_vocabulary import obo_cache, ControlledVocabulary
from .utils import add_metaclass, ensure_iterable, Mapping

from .xml import (
    id_maker, CVParam, UserParam,
    ParamGroupReference, _element, CVCollection,
    XMLWriterMixin)


class ChildTrackingMeta(type):
    def __new__(cls, name, parents, attrs):
        if not hasattr(cls, "_cache"):
            cls._cache = defaultdict(dict)
        new_type = type.__new__(cls, name, parents, attrs)
        ns = getattr(new_type, "component_namespace", None)
        cls._cache[ns][name] = new_type
        return new_type

    @classmethod
    def resolve_component(self, namespace, name):
        try:
            return self._cache[namespace][name]
        except KeyError:
            return self._cache[None][name]


class ReferentialIntegrityWarning(UserWarning):
    pass


class ReferentialIntegrityError(KeyError):
    pass


class SpecializedContextCache(OrderedDict):
    def __init__(self, type_name, missing_reference_is_error=False):
        super(SpecializedContextCache, self).__init__()
        self.type_name = type_name
        self.bijection = dict()
        self.preregistered = dict()
        self.missing_reference_is_error = missing_reference_is_error

    def __getitem__(self, key):
        try:
            item = dict.__getitem__(self, key)
            return item
        except KeyError:
            if key is None:
                if self.missing_reference_is_error:
                    raise ReferentialIntegrityError(
                        "A reference key for %s should not be \"None\"" % (self.type_name, ))
                else:
                    warnings.warn(
                        "A reference key for %s should not be \"None\"" % (self.type_name, ),
                        ReferentialIntegrityWarning,
                        stacklevel=3)
                    return None
            if self.missing_reference_is_error:
                raise ReferentialIntegrityError(key)
            else:
                warnings.warn(
                    "No reference was found for %r in %s" % (key, self.type_name),
                    ReferentialIntegrityWarning,
                    stacklevel=3)
            if isinstance(key, int):
                new_value = id_maker(self.type_name, key)
            else:
                new_value = str(key)
            self[key] = new_value
            return new_value

    def register(self, id):
        if id is None:
            return None
        if isinstance(id, int):
            value = id_maker(self.type_name, id)
        else:
            value = str(id)
        self[id] = value
        self.preregistered[id] = value
        return value

    def __setitem__(self, key, value):
        if key in self and key not in self.preregistered:
            warnings.warn(
                "Overwriting existing value for %r with %r in store %s" % (
                    key, value, self.type_name), ReferentialIntegrityWarning)
        super(SpecializedContextCache, self).__setitem__(key, value)
        self.bijection[value] = key

    def __repr__(self):
        return '%s\n%s' % (self.type_name, dict.__repr__(self))


class AmbiguousTermWarning(UserWarning):
    pass


class VocabularyResolver(object):
    warn_on_ambiguous_missing_units = True
    validate_units = True

    def __init__(self, vocabularies=None, vocabulary_resolver=None):
        if vocabularies is None:
            vocabularies = []
        if vocabulary_resolver is None:
            vocabulary_resolver = obo_cache
        self.vocabulary_resolver = vocabulary_resolver
        self.vocabularies = CVCollection(
            map(self._bind_vocabulary, vocabularies))

    def _bind_vocabulary(self, cv):
        cv.resolver = self.vocabulary_resolver
        return cv

    def get_vocabulary(self, id):
        for vocab in self.vocabularies:
            if vocab.id == id:
                return vocab
            try:
                if vocab.full_name == id:
                    return vocab
            except AttributeError:
                pass
        raise KeyError(id)

    def param_group_reference(self, id):
        return ParamGroupReference(id)

    def param(self, name, value=None, cv_ref=None, **kwargs):
        accession = kwargs.get("accession")
        if isinstance(name, CVParam):
            return name
        elif isinstance(name, ParamGroupReference):
            return name
        elif isinstance(name, (tuple, list)) and value is None:
            name, value = name
        elif isinstance(name, Mapping):
            mapping = dict(name)
            if len(mapping) == 1 and 'ref' in mapping:
                return self.param_group_reference(mapping['ref'])
            value = value or mapping.pop('value', None)
            accession = accession or mapping.pop("accession", None)
            cv_ref = cv_ref or mapping.pop("cv_ref", None) or mapping.pop("cvRef", None)
            unit_name = mapping.pop("unit_name", None) or mapping.pop("unitName", None)
            unit_accession = mapping.pop("unit_accession", None) or mapping.pop("unitAccession", None)
            unit_cv_ref = mapping.pop('unit_cv_ref', None) or mapping.pop('unitCvRef', None)
            name = mapping.pop('name', None)
            if name is None and accession is None:
                if len(mapping) == 1:
                    name, value = tuple(mapping.items())[0]
                else:
                    raise ValueError("Could not coerce parameter from %r" % (mapping,))
            else:
                kwargs.update({k: v for k, v in mapping.items()
                               if k not in (
                    "name", "value", "accession")})
            # case normalize unit information so that cvParam can detect them
            if unit_name is not None:
                kwargs.setdefault("unit_name", unit_name)
            if unit_accession is not None:
                kwargs.setdefault("unit_accession", unit_accession)
            if unit_cv_ref is not None:
                kwargs.setdefault("unit_cv_ref", unit_cv_ref)

        if name == 'ref' and value is not None and cv_ref is None and not kwargs:
            return self.param_group_reference(value)

        self._resolve_units(kwargs)
        if name is None and accession is None:
            raise ValueError("Could not coerce parameter from %r, %r, %r" % (name, value, kwargs))
        term = None
        if cv_ref is None:
            query = accession if accession is not None else name
            cv_ref, name, accession, term = self._resolve_cv_ref(query, name, accession)
        else:
            cv = self.get_vocabulary(cv_ref)
            query = accession if accession is not None else name
            term = cv[query]
        if term is not None:
            self._validate_units(term, kwargs, name)
        if cv_ref is None:
            user_param = UserParam(name=name, value=value, **kwargs)
            return user_param
        else:
            kwargs.setdefault("ref", cv_ref)
            kwargs.setdefault("accession", accession)
            return CVParam(name=name, value=value, **kwargs)

    def _resolve_cv_ref(self, query, name, accession):
        cv_ref = None
        term = None
        for cv in self.vocabularies:
            try:
                term = cv[query]
                name = term["name"]
                accession = term["id"]
                if cv_ref is not None:
                    raise ValueError(
                        "Resolutions exist for the term denoted by %r, found in %s and %s" % (
                            query, cv_ref, cv.id
                        ))
                cv_ref = cv.id
            except KeyError:
                continue
        return cv_ref, name, accession, term

    def _resolve_units(self, state):
        unit_name = state.get("unit_name")
        unit_accession = state.get("unit_accession")
        unit_ref = state.get('unit_cv_ref')
        if unit_name is not None or unit_accession is not None:
            if unit_accession is not None:
                unit_term, source = self.term(unit_accession, include_source=True)
                unit_name = unit_term.name
                unit_accession = unit_term.id
                unit_ref = source.id
            elif unit_name is not None:
                unit_term, source = self.term(unit_name, include_source=True)
                unit_name = unit_term.name
                unit_accession = unit_term.id
                unit_ref = source.id
            state['unit_name'] = unit_name
            state['unit_accession'] = unit_accession
            state['unit_cv_ref'] = unit_ref

    def _validate_units(self, term, state, name):
        has_units = term.get("has_units", [])
        if has_units:
            if len(has_units) == 1:
                provided_unit_accession = state.get("unit_accession")
                if provided_unit_accession is None:
                    try:
                        unit_term, unit_source = self.term(has_units[0].accession, include_source=True)
                        state['unit_accession'] = unit_term.id
                        state['unit_name'] = unit_term.name
                        state['unit_cv_ref'] = unit_source.id
                    except KeyError:
                        pass
                elif self.validate_units:
                    unit_term, unit_source = self.term(has_units[0].accession, include_source=True)
                    if (state['unit_accession'] != unit_term.id or state['unit_name'] != unit_term.name or
                            state['unit_cv_ref'] != unit_source.id):
                        warnings.warn("Provided unit for %r does not match the permitted unit (%r, %r, %r)" % (
                            name,
                            (state['unit_accession'], unit_term.id),
                            (state['unit_name'], unit_term.name),
                            (state['unit_cv_ref'], unit_source.id),
                        ), stacklevel=4)
            elif len(has_units) > 1:
                provided_unit_accession = state.get("unit_accession")
                if provided_unit_accession is None:
                    if self.warn_on_ambiguous_missing_units:
                        warnings.warn(
                            "Multiple unit options are possible for parameter %r but none were specified" % (
                                name),
                            AmbiguousTermWarning,
                            stacklevel=4
                        )
                    try:
                        unit_term, unit_source = self.term(has_units[0].accession, include_source=True)
                        state['unit_accession'] = unit_term.id
                        state['unit_name'] = unit_term.name
                        state['unit_cv_ref'] = unit_source.id
                    except KeyError:
                        pass
                elif self.validate_units:
                    for t in has_units:
                        unit_term, unit_source = self.term(t.accession, include_source=True)
                        if (state['unit_accession'] == unit_term.id and state['unit_name'] == unit_term.name and
                                state['unit_cv_ref'] == unit_source.id):
                            break
                    else:
                        warnings.warn(
                            "Provided unit for %r does not match any of the permitted units %r" % (
                                name,
                                ((state['unit_accession'], state['unit_name'], state['unit_cv_ref']),
                                 has_units)),
                            stacklevel=4
                        )

    def term(self, name, include_source=False):
        deferred = None
        for cv in self.vocabularies:
            try:
                term = cv[name]
                if term.get("is_obsolete", False):
                    deferred = term, cv
                    raise KeyError(name)
                if include_source:
                    return term, cv
                else:
                    return term
            except KeyError:
                pass
        else:
            if deferred:
                if include_source:
                    return deferred
                else:
                    return deferred[0]
            raise KeyError(name)

    def load_vocabularies(self):
        for vocab in self.vocabularies:
            vocab.load()

    def prepare_params(self, params):
        out = []
        for param in ensure_iterable(params):
            out.append(self.param(param))
        return out


class DocumentContext(dict, VocabularyResolver):

    def __init__(self, vocabularies=None, vocabulary_resolver=None, missing_reference_is_error=False):
        dict.__init__(self)
        VocabularyResolver.__init__(self, vocabularies, vocabulary_resolver)
        self.missing_reference_is_error = missing_reference_is_error

    def param_group_reference(self, id):
        # This is a inelegant, as ReferenceableParamGroup is not part document type
        # independent, and may not be consistent
        param_group_index = self['ReferenceableParamGroup']
        return ParamGroupReference(param_group_index[id])

    def __getitem__(self, key):
        if not isinstance(key, str):
            if isinstance(key, (type, ReprBorrowingPartial)):
                key = key.__name__
        return dict.__getitem__(self, key)

    def __setitem__(self, key, value):
        if not isinstance(key, str):
            if isinstance(key, (type, ReprBorrowingPartial)):
                key = key.__name__
        dict.__setitem__(self, key, value)

    def __missing__(self, key):
        if not isinstance(key, str):
            if isinstance(key, (type, ReprBorrowingPartial)):
                key = key.__name__
        self[key] = SpecializedContextCache(
            key, missing_reference_is_error=self.missing_reference_is_error)
        return self[key]


NullMap = DocumentContext()


class ReprBorrowingPartial(partial):
    """
    Create a partial instance that uses the wrapped callable's
    `__repr__` method instead of a generic partial
    """
    def __init__(self, func, *args, **kwargs):
        self._func = func
        update_wrapper(self, func)

    @property
    def type(self):
        return self._func

    def __repr__(self):
        return repr(self.func)

    def __getattr__(self, name):
        return getattr(self._func, name)

    def ensure(self, data):
        if not isinstance(data, self.type):
            return self(**data)
        else:
            if data.context is not self.context:
                raise ValueError("Cannot bind a component from another context")
        return data

    def ensure_all(self, objs):
        return [self.ensure(obj or {}) for obj in ensure_iterable(objs)]


class CallbackBindingPartial(ReprBorrowingPartial):

    callback = None

    def __call__(self, *args, **kwargs):
        result = super(CallbackBindingPartial, self).__call__(*args, **kwargs)
        if self.callback is not None:
            self.callback(result)
        return result


class ComponentDispatcherBase(object):
    """
    A container for a :class:`DocumentContext` which provides
    an automatically parameterized version of all :class:`ComponentBase`
    types which use this instance's context.

    Attributes
    ----------
    context : :class:`DocumentContext`
        The mapping responsible for managing the global
        state of all created components.
    """
    _component_partial_type = CallbackBindingPartial

    def __init__(self, context=None, vocabularies=None, vocabulary_resolver=None, component_namespace=None,
                 missing_reference_is_error=False):
        if vocabularies is None:
            vocabularies = []
        if context is None:
            context = DocumentContext(
                vocabularies=vocabularies, vocabulary_resolver=vocabulary_resolver,
                missing_reference_is_error=missing_reference_is_error)
        else:
            if vocabularies is not None:
                context.vocabularies.update(vocabularies)
        self.type_cache = dict()
        self.component_namespace = component_namespace
        self.context = context

    def _prepare_bind_arguments(self):
        return {'context': self.context}

    def _post_constructor(self, component):
        pass

    def _update_component_namespace(self, tp=None):
        return {}

    def _locate_component(self, name):
        try:
            tp = self.type_cache[name]
        except KeyError:
            tp_template = ChildTrackingMeta.resolve_component(self.component_namespace, name)
            new_tp = type(tp_template.__name__, (tp_template, ), self._update_component_namespace(tp_template))
            try:
                new_tp.__module__ = tp_template.__module__
            except AttributeError:
                pass
            try:
                new_tp.__qualname__ = tp_template.__qualname__
            except AttributeError:
                pass
            tp = self._component_partial_type(new_tp, **self._prepare_bind_arguments())
            self.type_cache[name] = tp
        return tp

    def _dispatch_component(self, name):
        tp = self._locate_component(name)
        tp.context = self.context
        tp.callback = self._post_constructor
        return tp

    def __getattr__(self, name):
        """
        Provide access to an automatically parameterized
        version of all :class:`ComponentBase` types which
        use this instance's context.

        Parameters
        ----------
        name : str
            Component Name

        Returns
        -------
        ReprBorrowingPartial
            A partially parameterized instance constructor for
            the :class:`ComponentBase` type requested.
        """
        if name.startswith("_"):
            raise AttributeError(name)
        tp = self._dispatch_component(name)
        return tp

    def ensure_component(self, data, tp):
        if isinstance(data, tp.type):
            if self.context is not data.context:
                raise ValueError("Cannot bind a component from another context")
            return data
        else:
            return tp(**data)

    def register(self, entity_type, id):
        """
        Pre-declare an entity in the document context. Ensures that
        a reference look up will be satisfied.

        Parameters
        ----------
        entity_type : str
            An entity type, either a tag name or a component name
        id : int
            The unique id number for the thing registered

        Returns
        -------
        str
            The constructed reference id
        """
        value = self.context[entity_type].register(id)
        return value

    @property
    def vocabularies(self):
        return self.context.vocabularies

    def load_vocabularies(self):
        self.context.load_vocabularies()

    def param(self, *args, **kwargs):
        return self.context.param(*args, **kwargs)

    def prepare_params(self, params):
        return self.context.prepare_params(params)

    def term(self, *args, **kwargs):
        return self.context.term(*args, **kwargs)

    def get_vocabulary(self, *args, **kwargs):
        return self.context.get_vocabulary(*args, **kwargs)


class XMLBindingDispatcherBase(ComponentDispatcherBase):

    def _post_constructor(self, component):
        component.writer = self.writer


# ------------------------------------------
# Base Component Definitions


@add_metaclass(ChildTrackingMeta)
class ComponentBase(object):
    """A base class for all parts of an XML document which
    describe structures composed of more than a single XML
    tag without any children. In addition to wrapping additional
    descriptive data, this type's metaclass is :class:`ChildTrackingMeta`
    which enables :class:`ComponentDispatcherBase` to automatically bind
    a :class:`DocumentContext` object to the `context` paramter of the
    constructor of dynamically generated wrappers.

    Forwards any missing attribute requests to :attr:`element` for resolution
    against's the XML tag's attributes.
    """

    # in any sane world, these would be initialized in a constructor but
    # would require substantial work to add the super-call to all constructors.
    _context_manager = None
    _is_open = False
    _after_queue = None
    requires_id = True
    writer = None

    def __init__(self, *args, **kwargs):
        pass

    def __getattr__(self, key):
        if key == "element":
            raise AttributeError(
                ("The `element` attribute failed to instantiate "
                 "or is being accessed too early."))
        try:
            return self.element.attrs[key]
        except KeyError:
            raise AttributeError(key)

    def after(self, callback):
        if self._after_queue is None:
            self._after_queue = []
        self._after_queue.append(callback)

    def write(self, xml_file):
        with self.begin(xml_file):
            pass

    def write_content(self, xml_file):
        raise NotImplementedError()

    def bind(self, writer):
        if isinstance(writer, XMLWriterMixin):
            writer = writer.writer
        self.writer = writer
        return self

    def is_bound(self):
        return self.writer is not None

    @contextmanager
    def begin(self, xml_file=None, with_id=None):
        if with_id is None:
            with_id = self.requires_id
        if xml_file is None:
            xml_file = getattr(self, "writer", None)
            if xml_file is None:
                raise ValueError("xml_file must be provided if this component is not bound to a writer!")
        self._is_open = True
        with self.element.begin(xml_file, with_id=with_id):
            self.write_content(xml_file)
            yield
            self._is_open = False
            if self._after_queue:
                for callback in self._after_queue:
                    callback(xml_file)

    def __enter__(self):
        if not self.is_bound():
            raise ValueError(
                "A component not bound to an XMLWriter cannot be used as a context manager directly. "
                "Call the `bind` method first with an XMLWriter")
        begun = self.begin()
        self._context_manager = begun
        # actually execute the code in `begin` now
        self._context_manager.__enter__()
        return self._context_manager

    def __exit__(self, exc_type, exc_value, traceback):
        # execute all instructions post-yield of begin
        self._context_manager.__exit__(exc_type, exc_value, traceback)
        self._context_manager = None

    def __call__(self, xml_file):
        self.write(xml_file)

    def __repr__(self):
        return "%s\n%s" % (
            self.element, "\n".join([
                "  %s: %r" % (k, v) for k, v in self.__dict__.items()
                if k not in ("context", "element") and not k.startswith("_")])
        )

    def prepare_params(self, params, **kwargs):
        if isinstance(params, Mapping):
            if ("name" not in params and "accession" not in params) and ("ref" not in params):
                params = list(params.items())
            else:
                params = [params]
        elif isinstance(params, (list, tuple)):
            params = list(params)
        else:
            params = list(ensure_iterable(params)) or []
        params.extend(kwargs.items())
        return params

    def has_param(self, query, params=None):
        if params is None:
            params = self.params
        query_param = self.context.term(query)
        for param in params:
            try:
                param = self.context.param(param)
                term = self.context.term(param.accession)
                if term.id == query_param.id:
                    return True
            except KeyError:
                continue
        return False

    def add_param(self, param):
        if isinstance(param, list):
            self.params.extend(param)
        else:
            self.params.append(param)
        return self

    def write_params(self, xml_file, params=None):
        if params is None:
            params = self.params
        params = self.prepare_params(params)
        user_params = []
        cv_params = []
        references = []
        for param in params:
            param = self.context.param(param)
            if isinstance(param, ParamGroupReference):
                references.append(param)
            elif isinstance(param, UserParam):
                user_params.append(param)
            else:
                cv_params.append(param)
        for param in (references + cv_params + user_params):
            param(xml_file)

    def element_attrs(self, **kwargs):
        for key, value in kwargs.items():
            if value is not None:
                self.element.attrs[key] = value

    @classmethod
    def ensure(cls, obj, **kwargs):
        if isinstance(obj, cls):
            return obj
        else:
            kwargs.update(obj)
            return cls(**kwargs)

    @classmethod
    def ensure_all(cls, objs, **kwargs):
        return [cls.ensure(obj or {}, **kwargs) for obj in ensure_iterable(objs)]


class ParameterContainer(ComponentBase):
    """An base class for a component whose only purpose
    is to contain one or more cv- or userParams.

    Attributes
    ----------
    context : DocumentContext
        The document metadata store
    element : lxml.etree.Element
        The XML tag object to be written
    params : list
        The list of parameters to include
    """
    requires_id = False

    def __init__(self, tag_name, params=None, element_args=None, context=NullMap, **kwargs):
        if element_args is None:
            element_args = dict()
        if params is None:
            params = []
        self.params = self.prepare_params(params, **kwargs)
        self.context = context
        self.element = _element(tag_name, **element_args)

    def write_content(self, xml_file):
        self.write_params(xml_file)


class IDParameterContainer(ParameterContainer):
    requires_id = True



from io import BytesIO
from psims.transform import mzml, utils

from psims.test.test_data import datafile


def test_mzml_pipe():
    buff = BytesIO()
    path = datafile("small.mzML")
    st = mzml.MzMLTransformer(path, buff)
    st.write()
    buff.seek(0)
    test_reader = mzml.MzMLParser(buff)
    ref_reader = st.reader
    ref_reader.reset()

    for ref_spec, test_spec in zip(ref_reader, test_reader):
        assert utils.differ(ref_spec, test_spec)

    ref_reader.reset()
    test_reader.reset()
    for ref_spec, test_spec in zip(ref_reader.iterfind("fileDescription"),
                                   test_reader.iterfind("fileDescription")):
        assert utils.differ(ref_spec, test_spec)


if __name__ == '__main__':
    test_mzml_pipe()



from .controlled_vocabulary import (
    ControlledVocabulary, OBOParser,
    load_psims, load_unimod,
    obo_cache, OBOCache)

from .mzml import (
    MzMLWriter, ARRAY_TYPES, compression_map,
    MZ_ARRAY, INTENSITY_ARRAY, CHARGE_ARRAY,
    components as mzml_components,
    default_cv_list as default_mzml_cv_list)

from .mzmlb import MzMLbWriter

from .mzid import (
    MzIdentMLWriter, default_cv_list as default_mzid_cv_list,
    components as mzid_components,)

from .version import version as __version__


from .utils import (checksum_file, TableStateMachine)

from . import compression


__all__ = [
    "ControlledVocabulary", "OBOParser", "load_psims", "load_unimod",
    "obo_cache", "OBOCache",

    "MzMLWriter", "ARRAY_TYPES", "compression_map", "MZ_ARRAY", "INTENSITY_ARRAY",
    "CHARGE_ARRAY", "mzml_components", "default_mzml_cv_list",

    "MzIdentMLWriter", "default_mzid_cv_list", "mzid_components",

    "compression",

    "TableStateMachine", "checksum_file", "__version__"
]



'''Machinery for interpreting XSD data type declarations in controlled vocabulary
file formats and mapping them to and from Python types.
'''
import re
import datetime
import warnings

from six import text_type

xsd_pattern = re.compile(r"(?:value-type:)?xsd\\?:([^\"]+)")


def non_negative_integer(value):
    x = int(value)
    if x < 0:
        raise TypeError("non_negative_integer cannot be negative. (%r)" % value)
    return x


def positive_integer(value):
    x = int(value)
    if x < 0:
        raise TypeError("positive_integer cannot be less than 1. (%r)" % value)
    return x


value_type_resolvers = {
    'int': int,
    'integer': int,
    'double': float,
    'float': float,
    'decimal': float,
    'string': text_type,
    "anyURI": text_type,
    'nonNegativeInteger': non_negative_integer,
    'boolean': bool,
    'positiveInteger': positive_integer,
    'dateTime': lambda x: datetime.datetime.strptime(x, '%Y-%m-%dT%H:%M:%S'),
    # TODO
    'date': str,
}


def parse_xsdtype(text):
    """Parse an XSD type definition to determine the appropriate Python type
    coercion function.

    Parameters
    ----------
    text : str
        The XSD type name

    Returns
    -------
    :class:`~.Callable`
    """
    match = xsd_pattern.search(text)
    if match:
        dtype_name = match.group(1).strip()
        try:
            dtype = value_type_resolvers[dtype_name]
        except KeyError:
            warnings.warn("Could not find a converter for XSD type %r" % (text, ))
            dtype = str
        return dtype


def obj_to_xsdtype(value):
    """Determine the appropriate XSD type from a Python object's
    type.

    Parameters
    ----------
    value : object
        The object whose appropriate XSD type to determine

    Returns
    -------
    :class:`str`:
        The XSD name for the type appropriate for `value`
    """
    if isinstance(value, bool):
        return "xsd:boolean"
    elif isinstance(value, int):
        return "xsd:int"
    elif isinstance(value, float):
        return "xsd:float"
    elif isinstance(value, text_type):
        return "xsd:string"
    else:
        return None



import numbers
import warnings

from collections import defaultdict

try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

import numpy as np

from psims.xml import XMLWriterMixin, XMLDocumentWriter
from psims.utils import TableStateMachine

from .components import (
    ComponentDispatcher, element,
    default_cv_list, MzML, InstrumentConfiguration, IndexedMzML)

from .binary_encoding import (
    encode_array, COMPRESSION_ZLIB,
    encoding_map, compression_map, dtype_to_encoding)

from .utils import ensure_iterable
from .index import IndexingStream

from .element_builder import ElementBuilder, ParamManagingProperty


MZ_ARRAY = 'm/z array'
INTENSITY_ARRAY = 'intensity array'
DEFAULT_INTENSITY_UNIT = "number of detector counts"
CHARGE_ARRAY = 'charge array'
TIME_ARRAY = "time array"
DEFAULT_TIME_UNIT = "minute"
NON_STANDARD_ARRAY = 'non-standard data array'

ARRAY_TYPES = [
    'm/z array',
    'intensity array',
    'charge array',
    'signal to noise array',
    'time array',
    'wavelength array',
    'flow rate array',
    'pressure array',
    'temperature array',
    'mean drift time array',
    'mean charge array',
    'resolution array',
    'baseline array'
]


class DocumentSection(ComponentDispatcher, XMLWriterMixin):

    def __init__(self, section, writer, parent_context, section_args=None, **kwargs):
        if section_args is None:
            section_args = dict()
        section_args.update(kwargs)
        super(DocumentSection, self).__init__(parent_context)
        self.section = section
        self.writer = writer
        self.section_args = section_args

    def __enter__(self):
        return self.begin()

    def __exit__(self, exc_type, exc_value, traceback):
        self.end(exc_type, exc_value, traceback)

    def begin(self):
        self.toplevel = element(self.writer, self.section, **self.section_args)
        self.toplevel.__enter__()
        return self

    def end(self, exc_type, exc_value, traceback):
        self.toplevel.__exit__(exc_type, exc_value, traceback)
        self.writer.flush()


class SpectrumListSection(DocumentSection):
    def __init__(self, writer, parent_context, section_args=None, **kwargs):
        super(SpectrumListSection, self).__init__(
            "spectrumList", writer, parent_context, section_args=section_args,
            **kwargs)
        self.section_args.setdefault("count", 0)
        data_processing_method = self.section_args.pop(
            "data_processing_method", None)
        try:
            self.section_args["defaultDataProcessingRef"] = self.context[
                "DataProcessing"][data_processing_method]
        except KeyError:
            try:
                self.section_args["defaultDataProcessingRef"] = list(
                    self.context["DataProcessing"].values())[0]
            except IndexError:
                warnings.warn(
                    "No Data Processing method found. mzML file may not be fully standard-compliant",
                    stacklevel=2)


class ChromatogramListSection(DocumentSection):
    def __init__(self, writer, parent_context, section_args=None, **kwargs):
        super(ChromatogramListSection, self).__init__(
            "chromatogramList", writer, parent_context,
            section_args=section_args, **kwargs)
        self.section_args.setdefault("count", 0)
        data_processing_method = self.section_args.pop(
            "data_processing_method", None)
        try:
            self.section_args["defaultDataProcessingRef"] = self.context[
                "DataProcessing"][data_processing_method]
        except KeyError:
            try:
                self.section_args["defaultDataProcessingRef"] = list(
                    self.context["DataProcessing"].values())[0]
            except IndexError:
                warnings.warn(
                    "No Data Processing method found. mzML file may not be fully standard-compliant",
                    stacklevel=2)


class RunSection(DocumentSection):
    """Describes a `<run>` tag. Implemented as a section to provide a more
    expressive API
    """
    def __init__(self, writer, parent_context, section_args=None, **kwargs):
        super(RunSection, self).__init__(
            "run", writer, parent_context, section_args=section_args, **kwargs)
        instrument_configuration_name = self.section_args.pop(
            "instrument_configuration", None)
        if instrument_configuration_name is not None:
            self.section_args["defaultInstrumentConfigurationRef"] = self.context[
                "InstrumentConfiguration"][instrument_configuration_name]
        source_file_name = self.section_args.pop("source_file", None)
        if source_file_name is not None:
            self.section_args["defaultSourceFileRef"] = self.context[
                "SourceFile"][source_file_name]
        sample_id = self.section_args.pop("sample", None)
        if sample_id is not None:
            self.section_args["sampleRef"] = self.context['Sample'][sample_id]


class IndexedmzMLSection(DocumentSection):
    def __init__(self, writer, parent_context, indexer, section_args=None, **kwargs):
        super(IndexedmzMLSection, self).__init__(
            'indexedmzML', writer, parent_context, section_args=section_args,
            **kwargs)
        self.toplevel = None
        self.inner = None
        self.indexer = indexer

    def begin(self):
        self.toplevel = element(self.writer, IndexedMzML())
        self.toplevel.__enter__()
        self.inner = element(self.writer, MzML(**self.section_args))
        self.inner.__enter__()

    def end(self, exc_type, exc_value, traceback):
        self.inner.__exit__(exc_type, exc_value, traceback)
        self.writer.flush()
        self.write_index()
        self.toplevel.__exit__(exc_type, exc_value, traceback)

    def write_index(self):
        self.indexer.to_xml(self)


class PlainMzMLWriter(ComponentDispatcher, XMLDocumentWriter):
    """A high level API for generating mzML XML files from simple Python objects.

    This class depends heavily on lxml's incremental file writing API which in turn
    depends heavily on context managers. Almost all logic is handled inside a context
    manager and in the context of a particular document. Since all operations assume
    that they have access to a universal identity map for each element in the document,
    that map is centralized in this class.

    MzMLWriter inherits from :class:`.ComponentDispatcher`, giving it a :attr:`context`
    attribute and access to all `Component` objects pre-bound to that context with attribute-access
    notation.

    Attributes
    ----------
    chromatogram_count : int
        A count of the number of chromatograms written
    spectrum_count : int
        A count of the number of spectra written
    """

    DEFAULT_TIME_UNIT = DEFAULT_TIME_UNIT
    DEFAULT_INTENSITY_UNIT = DEFAULT_INTENSITY_UNIT

    def __init__(self, outfile, close=False, vocabularies=None, missing_reference_is_error=False,
                 vocabulary_resolver=None, id=None, accession=None, **kwargs):
        if vocabularies is None:
            vocabularies = []
        vocabularies = list(default_cv_list) + list(vocabularies)
        ComponentDispatcher.__init__(
            self,
            vocabularies=vocabularies,
            vocabulary_resolver=vocabulary_resolver,
            missing_reference_is_error=missing_reference_is_error)
        XMLDocumentWriter.__init__(self, outfile, close, **kwargs)
        self.id = id
        self.accession = accession
        self.spectrum_count = 0
        self.chromatogram_count = 0
        self.default_instrument_configuration = None
        self.state_machine = TableStateMachine([
            ("start", ['controlled_vocabularies', ]),
            ("controlled_vocabularies", ['file_description', ]),
            ("file_description", ['reference_param_group_list', 'sample_list', 'software_list']),
            ("reference_param_group_list", ['sample_list', 'software_list']),
            ("sample_list", ['software_list', ]),
            ("software_list", ["scan_settings_list", 'instrument_configuration_list']),
            ("scan_settings_list", ['instrument_configuration_list', ]),
            ("instrument_configuration_list", ['data_processing_list']),
            ("data_processing_list", ['run']),
            ("run", ['spectrum_list', 'chromatogram_list']),
            ('spectrum_list', ['chromatogram_list']),
            ('chromatogram_list', [])
        ])

    def toplevel_tag(self):
        return MzML(id=self.id, accession=self.accession)

    def controlled_vocabularies(self):
        """Write out the `<cvList>` element and all its children,
        including both this format's default controlled vocabularies
        and those passed as arguments to this method.this

        This method requires writing to have begun.
        """
        self.state_machine.transition("controlled_vocabularies")
        super(PlainMzMLWriter, self).controlled_vocabularies()

    def software_list(self, software_list):
        """Writes the ``<softwareList>`` section of the document.

        .. note::
            List and descriptions of software used to acquire and/or process the
            data in this mzML file

        Parameters
        ----------
        software_list : list
            A list or other iterable of :class:`dict` or :class:`~.Software`-like objects
        """
        self.state_machine.transition("software_list")
        n = len(software_list)
        if n:
            software_list = [self.Software.ensure(sw) for sw in ensure_iterable(software_list)]
        self.SoftwareList(software_list).write(self)

    def file_description(self, file_contents=None, source_files=None, contacts=None):
        r"""Writes the ``<fileDescription>`` section of the document.

        .. note::
            Information pertaining to the entire mzML file (i.e. not specific
            to any part of the data set) is stored here.

        Parameters
        ----------
        file_contents : list, optional
            A list or other iterable of :class:`str`, :class:`dict`, or \*Param-types which will
            be placed in the ``<fileContent>`` element.
        source_files : list
            A list or other iterable of dict or :class:`~.SourceFile`-like objects
            to be placed in the ``<sourceFileList>`` element
        """
        self.state_machine.transition("file_description")
        fd = self.FileDescription(
            file_contents, [self.SourceFile.ensure(sf) for sf in ensure_iterable(source_files)],
            contacts=[self.Contact.ensure(c) for c in ensure_iterable(contacts)])
        fd.write(self.writer)

    def instrument_configuration_list(self, instrument_configurations):
        """Writes the ``<instrumentConfigurationList>`` section of the document.

        .. note::
            List and descriptions of instrument configurations. At least one instrument configuration MUST
            be specified, even if it is only to specify that the instrument is unknown. In that case, the
            "instrument model" term is used to indicate the unknown instrument in the instrumentConfiguration

        Parameters
        ----------
        instrument_configurations : list
            A list or other iterable of :class:`dict` or :class:`~.InstrumentConfiguration`-like
            objects
        """
        self.state_machine.transition("instrument_configuration_list")
        configs = [
            self.InstrumentConfiguration.ensure(ic) if not isinstance(
                ic, InstrumentConfiguration) else ic
            for ic in ensure_iterable(
                instrument_configurations)]
        self.InstrumentConfigurationList(configs).write(self)

    def data_processing_list(self, data_processing):
        """Writes the ``<dataProcessingList>`` section of the document.

        .. note::
            List and descriptions of data processing applied to this data

        Parameters
        ----------
        data_processing : list
            A list or other iterable of :class:`dict` or :class:`~.DataProcessing`-like
            objects

        """
        self.state_machine.transition("data_processing_list")
        methods = [
            self.DataProcessing.ensure(dp) for dp in ensure_iterable(data_processing)]
        self.DataProcessingList(methods).write(self)

    def reference_param_group_list(self, groups):
        """Writes the ``<referenceableParamGroupList>`` section of the document.

        Parameters
        ----------
        groups : list
            A list or other iterable of :class:`dict` or :class:`~.ReferenceableParamGroup`-like
            objects

        """
        self.state_machine.transition("reference_param_group_list")
        groups = [
            self.ReferenceableParamGroup.ensure(g) for g in ensure_iterable(groups)]
        self.ReferenceableParamGroupList(groups).write(self)

    def sample_list(self, samples):
        """Writes the ``<sampleList>`` section of the document

        Parameters
        ----------
        samples : list
            A list or other iterable of :class:`dict` or :class:`~.mzml.components.Sample`-like
            objects

        """
        self.state_machine.transition("sample_list")
        for i, sample in enumerate(ensure_iterable(samples)):
            if isinstance(sample, Mapping):
                sample_id = sample.get('id')
                sample_name = sample.get("name")

                if sample_id is None and sample_name is not None:
                    sample_id = "%s_%d_id" % (sample_name, i)
                elif sample_id is not None and sample_name is None:
                    sample_name = str(sample_id)
                elif sample_id is sample_name is None:
                    sample_id = "sample_%d_id" % (i,)
                    sample_name = "sample_%d" % (i,)
                sample['id'] = sample_id
                sample['name'] = sample_name

        sample_entries = self.Sample.ensure_all(samples)

        self.SampleList(sample_entries).write(self)

    def scan_settings_list(self, scan_settings):
        self.state_machine.transition("scan_settings_list")
        scan_settings = self.ScanSettings.ensure_all(scan_settings)
        self.ScanSettingsList(scan_settings).write(self)

    def run(self, id=None, instrument_configuration=None, source_file=None, start_time=None, sample=None):
        """Begins the `<run>` section of the document, describing a single
        sample run.

        Parameters
        ----------
        id : str, optional
            The unique identifier for this element
        instrument_configuration : str, optional
            The id string for the default `InstrumentConfiguration` for this
            sample
        source_file : str, optional
            The id string for the source file used to produce this data
        start_time : str, optional
            A string encoding the date and time the sample was acquired
        sample: str, optional
            The id string for the sample used to produce this data

        Returns
        -------
        RunSection
        """
        self.state_machine.transition("run")
        kwargs = {}
        if start_time is not None:
            kwargs['startTimeStamp'] = start_time
        if instrument_configuration is None:
            keys = list(self.context['InstrumentConfiguration'].keys())
            if keys:
                instrument_configuration = keys[0]
            else:
                instrument_configuration = None
        self.default_instrument_configuration = instrument_configuration
        return RunSection(
            self.writer, self.context, id=id,
            instrument_configuration=instrument_configuration,
            source_file=source_file,
            sample=sample, **kwargs)

    def spectrum_list(self, count, data_processing_method=None):
        self.state_machine.transition('spectrum_list')
        if data_processing_method is None:
            dp_map = self.context['DataProcessing']
            try:
                data_processing_method = list(dp_map.keys())[0]
            except IndexError:
                warnings.warn(
                    "No Data Processing method found. mzML file may not be fully standard-compliant",
                    stacklevel=2)
        return SpectrumListSection(
            self.writer, self.context, count=count,
            data_processing_method=data_processing_method)

    def chromatogram_list(self, count, data_processing_method=None):
        self.state_machine.transition('chromatogram_list')
        if data_processing_method is None:
            dp_map = self.context['DataProcessing']
            try:
                data_processing_method = list(dp_map.keys())[0]
            except IndexError:
                warnings.warn(
                    "No Data Processing method found. mzML file may not be fully standard-compliant",
                    stacklevel=2)
        return ChromatogramListSection(
            self.writer, self.context, count=count,
            data_processing_method=data_processing_method)

    def spectrum(self, mz_array=None, intensity_array=None, charge_array=None, id=None,
                 polarity='positive scan', centroided=True, precursor_information=None,
                 scan_start_time=None, params=None, compression=COMPRESSION_ZLIB,
                 encoding=None, other_arrays=None, scan_params=None, scan_window_list=None,
                 instrument_configuration_id=None, intensity_unit=DEFAULT_INTENSITY_UNIT):
        '''Create a new :class:`~.Spectrum` instance to be written.

        Parameters
        ----------
        mz_array: :class:`np.ndarray` of floats
            The m/z array of the spectrum
        intensity_array: :class:`np.ndarray` of floats
            The intensity array of the spectrum
        charge_array: :class:`np.ndarray`, optional
            The charge state array of the spectrum, optional.
        id: str
            The native ID of the spectrum.
        polarity: str or int, optional
            The polarity of the spectrum. If an integer, the sign of
            the integer is used, otherwise it is interpreted as a cvParam
        centroided: bool, optional
            Whether the spectrum is continuous or discretized by peak picking.
            Defaults to :const:`True`.
        precursor_information: dict or :class:`PrecursorBuilder`, optional
            The precursor ion description. Will be passed to :meth:`_prepare_precursor_list`.
            The structure of this object should either be formatted as arguments to
            :meth:`precursor_builder`, or a :class:`PrecursorBuilder` instance populated
            with information.
        scan_start_time: float, optional
            The scan start time, in minutes
        params: list, optional
            The parameters of the `spectrum`
        compression: str, optional
            The compression type name to use. Defaults to `COMPRESSION_ZLIB`.
        encoding: dict, optional
            A mapping from array name to NumPy data types.
        other_arrays: dict, optional
            A mapping of array names to additional data arrays
        scan_params: list, optional
            A list of cvParams for the `scan` of this `spectrum`
        scan_window_list: list, optional
            A list of scan windows specified as pairs of m/z intervals
        instrument_configuration_id: str, optional
            The `id` of the `instrumentConfiguration` to associate with this spectrum
            if not the default one.

        Returns
        -------
        :class:`~.Spectrum`
        '''
        self.state_machine.expects_state("spectrum_list")
        if encoding is None:
            encoding = {MZ_ARRAY: np.float64}
        if params is None:
            params = []
        else:
            params = list(params)
        if scan_params is None:
            scan_params = []
        else:
            scan_params = list(scan_params)
        if other_arrays is None:
            other_arrays = []
        if scan_window_list is None:
            scan_window_list = []
        else:
            scan_window_list = list(scan_window_list)

        if isinstance(encoding, Mapping):
            encoding = defaultdict(lambda: np.float32, encoding)
        else:
            # create new variable to capture in closure
            _encoding = encoding
            encoding = defaultdict(lambda: _encoding)
        if polarity is not None:
            if isinstance(polarity, int):
                if polarity > 0:
                    polarity = 'positive scan'
                elif polarity < 0:
                    polarity = 'negative scan'
                else:
                    polarity = None
            elif 'positive' in polarity:
                polarity = 'positive scan'
            elif 'negative' in polarity:
                polarity = 'negative scan'
            else:
                polarity = None

            if polarity not in params and polarity is not None:
                params.append(polarity)

        if centroided:
            peak_mode = "centroid spectrum"
        else:
            peak_mode = 'profile spectrum'
        params.append(peak_mode)

        array_list = []
        default_array_length = len(mz_array) if mz_array is not None else 0
        if mz_array is not None:
            mz_array_tag = self._prepare_array(
                mz_array, encoding=encoding[MZ_ARRAY], compression=compression, array_type=MZ_ARRAY, scope='spectrum')
            array_list.append(mz_array_tag)

        if intensity_array is not None:
            intensity_array_tag = self._prepare_array(
                intensity_array, encoding=encoding[INTENSITY_ARRAY], compression=compression,
                array_type={"name": INTENSITY_ARRAY, "unit_name": intensity_unit}, scope='spectrum')
            array_list.append(intensity_array_tag)

        if charge_array is not None:
            charge_array_tag = self._prepare_array(
                charge_array, encoding=encoding[CHARGE_ARRAY], compression=compression,
                array_type=CHARGE_ARRAY, scope='spectrum')
            array_list.append(charge_array_tag)
        for array_type, array in other_arrays:
            if array_type is None:
                raise ValueError("array type can't be None")
            array_tag = self._prepare_array(
                array, encoding=encoding[array_type], compression=compression, array_type=array_type,
                default_array_length=default_array_length, scope='spectrum')
            array_list.append(array_tag)
        array_list_tag = self.BinaryDataArrayList(array_list)

        if precursor_information is not None:
            precursor_list = self._prepare_precursor_list(
                precursor_information, intensity_unit=intensity_unit)
        else:
            precursor_list = None

        if scan_start_time is not None:
            if isinstance(scan_start_time, numbers.Number):
                scan_params.append({"name": "scan start time",
                                    "value": scan_start_time,
                                    "unitName": DEFAULT_TIME_UNIT})
            else:
                scan_params.append(scan_start_time)
        # The spec says this is optional, but the validator calls this a must
        # if self.default_instrument_configuration == instrument_configuration_id:
        #     instrument_configuration_id = None
        scan = self.Scan(scan_window_list=scan_window_list, params=scan_params,
                         instrument_configuration_ref=instrument_configuration_id)
        scan_list = self.ScanList([scan], params=["no combination"])

        index = self.spectrum_count
        self.spectrum_count += 1
        spectrum = self.Spectrum(
            index, array_list_tag, scan_list=scan_list, params=params, id=id,
            default_array_length=default_array_length,
            precursor_list=precursor_list)
        return spectrum

    def write_spectrum(self, mz_array=None, intensity_array=None, charge_array=None, id=None,
                       polarity='positive scan', centroided=True, precursor_information=None,
                       scan_start_time=None, params=None, compression=COMPRESSION_ZLIB,
                       encoding=None, other_arrays=None, scan_params=None, scan_window_list=None,
                       instrument_configuration_id=None, intensity_unit=DEFAULT_INTENSITY_UNIT):
        '''Write a :class:`~.Spectrum` with the provided data.

        Parameters
        ----------
        mz_array: :class:`np.ndarray` of floats
            The m/z array of the spectrum
        intensity_array: :class:`np.ndarray` of floats
            The intensity array of the spectrum
        charge_array: :class:`np.ndarray`, optional
            The charge state array of the spectrum, optional.
        id: str
            The native ID of the spectrum.
        polarity: str or int, optional
            The polarity of the spectrum. If an integer, the sign of
            the integer is used, otherwise it is interpreted as a cvParam
        centroided: bool, optional
            Whether the spectrum is continuous or discretized by peak picking.
            Defaults to :const:`True`.
        precursor_information: dict or :class:`PrecursorBuilder`, optional
            The precursor ion description. Will be passed to :meth:`_prepare_precursor_list`.
            The structure of this object should either be formatted as arguments to
            :meth:`precursor_builder`, or a :class:`PrecursorBuilder` instance populated
            with information.
        scan_start_time: float, optional
            The scan start time, in minutes
        params: list, optional
            The parameters of the `spectrum`
        compression: str, optional
            The compression type name to use. Defaults to `COMPRESSION_ZLIB`.
        encoding: dict, optional
            A mapping from array name to NumPy data types.
        other_arrays: dict, optional
            A mapping of array names to additional data arrays
        scan_params: list, optional
            A list of cvParams for the `scan` of this `spectrum`
        scan_window_list: list, optional
            A list of scan windows specified as pairs of m/z intervals
        instrument_configuration_id: str, optional
            The `id` of the `instrumentConfiguration` to associate with this spectrum
            if not the default one.

        See Also
        --------
        :meth:`spectrum`
        '''
        spectrum = self.spectrum(
            mz_array=mz_array, intensity_array=intensity_array, charge_array=charge_array,
            id=id, polarity=polarity, centroided=centroided, precursor_information=precursor_information,
            scan_start_time=scan_start_time, params=params, compression=compression,
            encoding=encoding, other_arrays=other_arrays, scan_params=scan_params,
            scan_window_list=scan_window_list,
            instrument_configuration_id=instrument_configuration_id,
            intensity_unit=intensity_unit)
        spectrum.write(self.writer)

    def chromatogram(self, time_array, intensity_array, id=None,
                     chromatogram_type="selected ion current",
                     precursor_information=None, params=None,
                     compression=COMPRESSION_ZLIB, encoding=32, other_arrays=None,
                     intensity_unit=DEFAULT_INTENSITY_UNIT, time_unit=DEFAULT_TIME_UNIT):
        self.state_machine.expects_state("chromatogram_list")
        if params is None:
            params = []
        else:
            params = list(params)

        if isinstance(encoding, Mapping):
            encoding = defaultdict(lambda: np.float32, encoding)
        else:
            # create new variable to capture in closure
            _encoding = encoding
            encoding = defaultdict(lambda: _encoding)

        if other_arrays is None:
            other_arrays = []
        array_list = []

        if precursor_information is not None:
            precursor = self._prepare_precursor_list(
                precursor_information, intensity_unit=intensity_unit)[0]
        else:
            precursor = None

        default_array_length = len(time_array) if time_array is not None else 0
        if time_array is not None:
            time_array_tag = self._prepare_array(
                time_array, encoding=encoding[TIME_ARRAY], compression=compression,
                array_type={"name": TIME_ARRAY, "unit_name": time_unit}, scope='chromatogram')
            array_list.append(time_array_tag)

        if intensity_array is not None:
            intensity_array_tag = self._prepare_array(
                intensity_array, encoding=encoding[INTENSITY_ARRAY], compression=compression,
                array_type={"name": INTENSITY_ARRAY, "unit_name": intensity_unit}, scope='chromatogram')
            array_list.append(intensity_array_tag)

        for array_type, array in other_arrays:
            array_tag = self._prepare_array(
                array, encoding=encoding[array_type], compression=compression, array_type=array_type,
                default_array_length=default_array_length, scope='chromatogram')
            array_list.append(array_tag)
        params.append(chromatogram_type)
        array_list_tag = self.BinaryDataArrayList(array_list)
        index = self.chromatogram_count
        self.chromatogram_count += 1
        chromatogram = self.Chromatogram(
            index=index, binary_data_list=array_list_tag,
            precursor=precursor,
            default_array_length=default_array_length,
            id=id, params=params)
        return chromatogram

    def write_chromatogram(self, time_array, intensity_array, id=None,
                           chromatogram_type="selected ion current",
                           precursor_information=None, params=None,
                           compression=COMPRESSION_ZLIB, encoding=32, other_arrays=None,
                           intensity_unit=DEFAULT_INTENSITY_UNIT, time_unit=DEFAULT_TIME_UNIT):
        chromatogram = self.chromatogram(
            time_array=time_array, intensity_array=intensity_array, id=id,
            chromatogram_type=chromatogram_type, precursor_information=precursor_information,
            params=params, compression=compression, encoding=encoding,
            other_arrays=other_arrays, intensity_unit=intensity_unit, time_unit=time_unit)
        chromatogram.write(self.writer)

    def _prepare_array(self, array, encoding=32, compression=COMPRESSION_ZLIB,
                       array_type=None, default_array_length=None, scope=None):
        if isinstance(encoding, numbers.Number):
            _encoding = int(encoding)
        else:
            _encoding = encoding
        dtype = encoding_map[_encoding]
        array = np.array(array, dtype=dtype)
        encoded_binary = encode_array(
            array, compression=compression, dtype=dtype)
        binary = self.Binary(encoded_binary)
        if default_array_length is not None and len(array) != default_array_length:
            override_length = True
        else:
            override_length = False
        params = []
        if array_type is not None:
            params.append(array_type)
            if isinstance(array_type, Mapping):
                array_type_ = array_type['name']
            else:
                array_type_ = array_type
            if array_type_ not in ARRAY_TYPES:
                params.append(NON_STANDARD_ARRAY)
        params.append(compression_map[compression])
        params.append(dtype_to_encoding[dtype])
        encoded_length = len(encoded_binary)
        return self.BinaryDataArray(
            binary, encoded_length,
            array_length=(len(array) if override_length else None),
            params=params)

    def _prepare_precursor_list(self, precursors, intensity_unit=DEFAULT_INTENSITY_UNIT):
        if isinstance(precursors, self.PrecursorList.type):
            return precursors
        elif isinstance(precursors, (dict)):
            precursors = self.PrecursorList([self._prepare_precursor_information(
                intensity_unit=intensity_unit, **precursors)])
        elif isinstance(precursors, PrecursorBuilder):
            precursors = self.PrecursorList([self._prepare_precursor_information(
                precursors,
                intensity_unit=intensity_unit)])
        else:
            packaged = []
            for p in ensure_iterable(precursors):
                if isinstance(p, self.Precursor.type):
                    packaged.append(p)
                elif isinstance(p, dict):
                    packaged.append(
                        self._prepare_precursor_information(
                            intensity_unit=intensity_unit, **p))
                elif isinstance(p, PrecursorBuilder):
                    packaged.append(
                        self._prepare_precursor_information(
                            p, intensity_unit=intensity_unit))
            precursors = self.PrecursorList(packaged)
        return precursors

    def _prepare_precursor_information(self, mz=None, intensity=None, charge=None, spectrum_reference=None, activation=None,
                                       isolation_window_args=None, params=None,
                                       intensity_unit=DEFAULT_INTENSITY_UNIT, scan_id=None, external_spectrum_id=None,
                                       source_file_reference=None):
        '''Prepare a :class:`Precursor` element from disparate data structures.

        Parameters
        ----------
        mz: float, optional
            The m/z of the first selected ion
        intensity: float, optional
            The intensity of the first selected ion
        charge: int, optional
            The charge state of the first seelcted ion
        spectrum_reference: str, optional
            The `id` of the prescursor `<spectrum>` for this precursor
        activation: dict, optional
            Parameters forwarded to :meth:`PrecursorBuilder.activation`
        isolation_window_args: tuple, list, or dict, optional
            Parameters forwarded to :meth:PrecursorBuilder.isolation_window`,
            tuple or list values are converted into :class:`dict` of the correct
            structure.
        params: list, optional
            The cv-params of the first selected ion
        intensity_unit: str
            The intensity unit of the first selected ion
        scan_id: str, optional
            An alias for `spectrum_reference`
        external_spectrum_id: str, optional
            The `externalSpectrumID` attribute of the precursor
        source_file_reference: str, optional
            The `sourceFileRef` attribute of the precursor

        Returns
        -------
        :class:`~.Precursor`
        '''
        if isinstance(mz, PrecursorBuilder):
            return self.Precursor(**mz.pack())
        if scan_id is not None:
            spectrum_reference = scan_id
        if params is None:
            params = []
        if activation:
            activation = self.Activation(activation)
        if any((mz, intensity, charge)):
            ion = self.SelectedIon(mz, intensity, charge, params=params)
            ion_list = self.SelectedIonList([ion])
        else:
            ion_list = None
        if isolation_window_args:
            isolation_window_tag = self.IsolationWindow(**isolation_window_args)
        else:
            isolation_window_tag = None
        precursor = self.Precursor(
            ion_list,
            activation=activation,
            isolation_window=isolation_window_tag,
            spectrum_reference=spectrum_reference)
        return precursor

    def precursor_builder(self, mz=None, intensity=None, charge=None, spectrum_reference=None, activation=None,
                          isolation_window_args=None, params=None,
                          intensity_unit=DEFAULT_INTENSITY_UNIT, scan_id=None,
                          external_spectrum_id=None,
                          source_file_reference=None):
        '''Create a :class:`PrecursorBuilder`, an object to help populate the precursor information
        data structure.

        The helper object should be used to incrementally populate the precursor information passed
        to :meth:`spectrum` or :meth:`write_spectrum`'s `precursor_information` argument.

        Parameters
        ----------
        mz: float, optional
            The m/z of the first selected ion
        intensity: float, optional
            The intensity of the first selected ion
        charge: int, optional
            The charge state of the first selected ion
        spectrum_reference: str, optional
            The `id` of the prescursor `<spectrum>` for this precursor, mapped through the
            document context.
        activation: dict or list, optional
            Parameters forwarded to :meth:`PrecursorBuilder.activation`. This should be a dictionary
            with a key "params" and a list of :class:`~.CVParam` coerce-able values, with additional
            optional keys naming other :class:`~.CVParam` coerce-able values.
        isolation_window_args: tuple, list, or dict, optional
            Parameters forwarded to :meth:PrecursorBuilder.isolation_window`,
            tuple or list of three values are converted into :class:`dict` of the correct
            structure. The expected keys are "lower", the lower m/z offset, "target", the center m/z,
            and "upper", the upper m/z offset.
        params: list, optional
            The cv- and user-params of the first selected ion, in addition to `mz`, `intensity`,
            `charge`.
        intensity_unit: str
            The intensity unit of the first selected ion, to be specified with `intensity`
        scan_id: str, optional
            An alias for `spectrum_reference`
        external_spectrum_id: str, optional
            The `externalSpectrumID` attribute of the precursor
        source_file_reference: str, optional
            The `sourceFileRef` attribute of the precursor

        Returns
        -------
        :class:`PrecursorBuilder`
        '''
        if scan_id is None:
            spectrum_reference = scan_id
        inst = PrecursorBuilder(
            self, spectrum_reference=spectrum_reference,
            external_spectrum_id=external_spectrum_id)
        if mz is not None or intensity is not None or charge is not None or params is not None:
            inst.selected_ion(
                mz=mz, intensity=intensity, charge=charge,
                intensity_unit=intensity_unit, params=params)
        if isolation_window_args is None:
            if isinstance(isolation_window_args, (tuple, list)):
                isolation_window_args = {
                    "lower": isolation_window_args[0],
                    "target": isolation_window_args[1],
                    "upper": isolation_window_args[2]}
            inst.isolation_window(isolation_window_args)
        if activation is not None:
            inst.activation(activation)
        return  inst


class SelectedIonBuilder(ElementBuilder):
    mz = ParamManagingProperty('selected_ion_mz', 0.0, aliases=['mz'])
    charge = ParamManagingProperty('charge')
    intensity = ParamManagingProperty('intensity', 0.0)
    intensity_unit = ParamManagingProperty(
        'intensity_unit', DEFAULT_INTENSITY_UNIT)


class IsolationWindowBuilder(ElementBuilder):
    lower = ParamManagingProperty('lower')
    target = ParamManagingProperty('target')
    upper = ParamManagingProperty('upper')


class ActivationBuilder(ElementBuilder):
    pass


class PrecursorBuilder(ElementBuilder):
    def __init__(self, source, binding=None, params=None, **kwargs):
        super(PrecursorBuilder, self).__init__(
            source, binding, params, **kwargs)

    selected_ion_list = ParamManagingProperty("selected_ion_list", list)
    _isolation_window = ParamManagingProperty("isolation_window")
    _activation = ParamManagingProperty('activation')

    spectrum_reference = ParamManagingProperty('spectrum_reference')
    source_file_reference = ParamManagingProperty('source_file_reference')
    external_spectrum_id = ParamManagingProperty('external_spectrum_id', aliases=['scan_id'])

    def selected_ion(self, binding=None, **kwargs):
        sib = SelectedIonBuilder(self.source, binding=binding, **kwargs)
        self.selected_ion_list.append(sib)
        return sib

    def isolation_window(self, binding=None, **kwargs):
        self._isolation_window = IsolationWindowBuilder(
            self.source, binding=binding, **kwargs)
        return self._isolation_window

    def activation(self, binding=None, **kwargs):
        self._activation = ActivationBuilder(
            self.source, binding=binding, **kwargs)
        return self._activation


class IndexedMzMLWriter(PlainMzMLWriter):
    def __init__(self, outfile, close=False, vocabularies=None, missing_reference_is_error=False,
                 vocabulary_resolver=None, id=None, accession=None, **kwargs):
        outfile = IndexingStream(outfile)
        super(IndexedMzMLWriter, self).__init__(
            outfile, close, vocabularies, missing_reference_is_error, vocabulary_resolver,
            id, accession, **kwargs)
        self.index_builder = outfile

    def toplevel_tag(self):
        return IndexedmzMLSection(
            self.writer, self.context, id=self.id, accession=self.accession,
            indexer=self.index_builder)

    def format(self, *args, **kwargs):
        return


MzMLWriter = IndexedMzMLWriter
# MzMLWriter = PlainMzMLWriter



from pyteomics import mzml

from psims import MzMLWriter, MzMLbWriter
from psims.utils import ensure_iterable

from .utils import TransformerBase


class MzMLParser(mzml.MzML):

    def _handle_param(self, element, **kwargs):
        try:
            element.attrib["value"]
        except KeyError:
            element.attrib["value"] = ""
        return super(MzMLParser, self)._handle_param(element, **kwargs)

    def reset(self):
        super(MzMLParser, self).reset()
        self.seek(0)


def identity(x):
    return x


class MzMLTransformer(TransformerBase):
    """Reads an mzML file stream from :attr:`input_stream`, copying its metadata
    to :attr:`output_stream`, and then copies its spectra, applying :attr:`transform`
    to each spectrum object as it goes.

    If :attr:`sort_by_by_scan_time` is :const:`True`, then prior to writing spectra,
    a first pass will be made over the mzML file and the spectra will be written out
    ordered by ``MS:1000016:"scan start time"``.

    Attributes
    ----------
    input_stream : file-like
        A byte stream from an mzML format data buffer
    output_stream : file-like
        A writable binary stream to copy the contents of :attr:`input_stream` into
    sort_by_scan_time : :class:`bool`
        Whether or not to sort spectra by scan time prior to writing
    transform : :class:`Callable`, optional
        A function to call on each spectrum, passed as a :class:`dict` object as
        read by :class:`pyteomics.mzml.MzML`.
    transform_description : :class:`str`
        A description of the transformation to include in the written metadata
    """

    def __init__(self, input_stream, output_stream, transform=None, transform_description=None,
                 sort_by_scan_time=False):
        if transform is None:
            transform = identity
        self.input_stream = input_stream
        self.output_stream = output_stream
        self.transform = transform
        self.transform_description = transform_description
        self.sort_by_scan_time = sort_by_scan_time
        self.reader = MzMLParser(input_stream, iterative=True)
        self.writer = MzMLWriter(output_stream)
        self.psims_cv = self.writer.get_vocabulary('PSI-MS').vocabulary

    def format_referenceable_param_groups(self):
        self.reader.reset()
        try:
            param_list = next(self.reader.iterfind("referenceableParamGroupList", recursive=True, retrive_refs=False))
            param_groups = ensure_iterable(param_list.get("referenceableParamGroup", []))
        except StopIteration:
            param_groups = []
        return [self.writer.ReferenceableParamGroup.ensure(d) for d in param_groups]

    def format_instrument_configuration(self):
        self.reader.reset()
        configuration_list = next(self.reader.iterfind("instrumentConfigurationList", recursive=True))
        configurations = []
        for config_dict in configuration_list.get("instrumentConfiguration", []):
            components = []
            for key, members in config_dict.pop('componentList', {}).items():
                if key not in ("source", "analyzer", "detector"):
                    continue
                if key == 'source':
                    components.extend(self.writer.Source.ensure(m) for m in members)
                elif key == "analyzer":
                    components.extend(self.writer.Analyzer.ensure(m) for m in members)
                elif key == 'detector':
                    components.extend(self.writer.Detector.ensure(m) for m in members)
            components.sort(key=lambda x: x.order)
            software_reference = config_dict.pop("softwareRef", {}).get("ref")
            configuration = self.writer.InstrumentConfiguration(
                component_list=components, software_reference=software_reference, **config_dict)
            configurations.append(configuration)
        return configurations

    def format_data_processing(self):
        self.reader.reset()
        dpl = next(self.reader.iterfind("dataProcessingList", recursive=True))
        data_processing = []
        for dp_dict in dpl.get("dataProcessing", []):
            methods = []
            for pm in dp_dict.pop("processingMethod", []):
                pm['software_reference'] = pm.pop("softwareRef")
                methods.append(self.writer.ProcessingMethod.ensure(pm))
            dp_dict['processing_methods'] = methods
            data_processing.append(self.writer.DataProcessing.ensure(dp_dict))
        return data_processing

    def copy_metadata(self):
        self.reader.reset()
        file_description = next(self.reader.iterfind("fileDescription"))
        source_files = file_description.get("sourceFileList").get('sourceFile')
        self.writer.file_description(file_description.get("fileContent", {}).items(), source_files)

        param_groups = self.format_referenceable_param_groups()
        if param_groups:
            self.writer.reference_param_group_list(param_groups)

        self.reader.reset()
        software_list = next(self.reader.iterfind("softwareList"))
        software_list = software_list.get("software", [])
        software_list.append(self._make_software())
        self.writer.software_list(software_list)

        configurations = self.format_instrument_configuration()
        self.writer.instrument_configuration_list(configurations)

        # include transformation description here
        data_processing = self.format_data_processing()
        data_processing.append(self._make_data_processing_entry())
        self.writer.data_processing_list(data_processing)

    def _make_software(self):
        description = {
            "id": "psims-MzMLTransformer",
            "params": [
                self.writer.param("python-psims"),
            ]
        }
        return description

    def _make_data_processing_entry(self):
        description = {
            "id": "psims-MzMLTransformer-processing",
            "processing_methods": [
                {
                    "order": 1,
                    "software_reference": "psims-MzMLTransformer",
                    "params": ([self.transform_description] if self.transform_description else []
                               ) + ['conversion to mzML'],
                }
            ]
        }
        return description

    def format_scan(self, scan):
        scan_params = []
        scan_window_list = []
        scan_start_time = None

        temp = scan.copy()

        for key, value in list(temp.items()):
            if not hasattr(key, 'accession'):
                continue
            accession = key.accession
            if accession == '' or accession is None:
                scan_params.append({key: value})
                if hasattr(value, 'unit_info'):
                    scan_params[-1]['unit_name'] = value.unit_info
                temp.pop(key)
                continue
            term = self.psims_cv[accession]
            if term.is_of_type("scan attribute"):
                if term.name == 'scan start time':
                    scan_start_time = {
                        "name": term.id, "value": value, "unit_name": getattr(value, 'unit_info', None)
                    }
                else:
                    scan_params.append({"name": term.id, "value": value})
                    if hasattr(value, 'unit_info'):
                        scan_params[-1]['unit_name'] = value.unit_info
                temp.pop(key)
        temp = temp.get('scanWindowList', {}).get('scanWindow', [{}])[0].copy()
        for key, value in list(temp.items()):
            if not hasattr(key, 'accession'):
                continue
            accession = key.accession
            term = self.psims_cv[accession]
            if term.is_of_type("selection window attribute"):
                scan_window_list.append(
                    {"name": term.id, "value": value})
                if hasattr(value, 'unit_info'):
                    scan_window_list[-1]['unit_name'] = value.unit_info
                temp.pop(key)

        scan_window_list.sort(key=lambda x: x['value'])
        if len(scan_window_list) % 2 == 0:
            windows = []
            i = 0
            n = len(scan_window_list)
            while i < n:
                lo = scan_window_list[i]
                hi = scan_window_list[i + 1]
                windows.append((lo['value'], hi['value']))
                i += 2
            scan_window_list = windows
        else:
            scan_window_list = []

        return scan_start_time, scan_params, scan_window_list

    def format_spectrum(self, spectrum):
        spec_data = dict()
        spec_data["mz_array"] = spectrum.pop("m/z array", None)
        spec_data["intensity_array"] = spectrum.pop("intensity array", None)
        spec_data["charge_array"] = spectrum.pop("charge array", None)

        spec_data['encoding'] = {
            "m/z array": spec_data["mz_array"].dtype.type if spec_data.get('mz_array') is not None else None,
            "intensity array": spec_data["intensity_array"].dtype.type if spec_data.get(
                'intensity_array') is not None else None,
            "charge array": spec_data["charge_array"].dtype.type if spec_data.get(
                'charge_array') is not None else None
        }

        spec_data['id'] = spectrum["id"]
        params = []

        if "positive scan" in spectrum:
            spec_data['polarity'] = 1
        elif "negative scan" in spectrum:
            spec_data['polarity'] = -1
        else:
            spec_data['polarity'] = None

        temp = spectrum.copy()
        for key, value in list(temp.items()):
            if not hasattr(key, 'accession'):
                continue
            accession = key.accession
            if accession == '' or accession is None:
                params.append({key: value})
                if hasattr(value, 'unit_info'):
                    params[-1]['unit_name'] = value.unit_info
                temp.pop(key)
            term = self.psims_cv[accession]
            if term.is_of_type("spectrum representation"):
                spec_data["centroided"] = term.id == "MS:1000127"
                temp.pop(key)
            elif term.is_of_type("spectrum property") or term.is_of_type("spectrum attribute"):
                params.append({"name": term.id, "value": value})
                if hasattr(value, 'unit_info'):
                    params[-1]['unit_name'] = value.unit_info
                temp.pop(key)

        spec_data["scan_start_time"], spec_data['scan_params'], spec_data["scan_window_list"] = self.format_scan(
            spectrum.get("scanList", {}).get('scan', [{}])[0])

        spec_data['params'] = params

        precursors = spectrum.get("precursorList", {}).get("precursor")
        if precursors:
            precursor_list = []
            for prec in precursors:
                precursor_information = {}
                precursor_information['scan_id'] = prec.get("spectrumRef")
                ion = prec['selectedIonList'].get("selectedIon")[0]
                for key, value in list(ion.items()):
                    term = self.psims_cv[key]
                    if term.id == "MS:1000744":
                        precursor_information['mz'] = value
                        ion.pop(key)
                    elif term.id == "MS:1000042":
                        precursor_information['intensity'] = value
                        ion.pop(key)
                    elif term.id in ("MS:1000041", "MS:1000633"):
                        precursor_information['charge'] = value
                        ion.pop(key)
                precursor_information.setdefault("intensity", None)
                precursor_information.setdefault("charge", None)
                precursor_information['params'] = ion.items()
                precursor_information['activation'] = (prec.get('activation', {}).items())
                precursor_information['isolation_window_args'] = prec.get("isolationWindow", None)
                precursor_list.append(precursor_information)

        else:
            precursor_list = None
        spec_data['precursor_information'] = precursor_list
        # attempt to find the instrumentConfiguration id to reference
        try:
            spec_data['instrument_configuration_id'] = spectrum.get(
                "scanList", {}).get("scan")[0].get("instrumentConfigurationRef")
        except IndexError:
            pass
        return spec_data

    def iterspectrum(self):
        self.reader.reset()
        if self.sort_by_scan_time:
            time_map = dict()
            self.log("Building Scan Time Map")
            for spectrum in self.reader.iterfind("spectrum"):
                time = self.reader._get_time(spectrum)
                time_map[spectrum['id']] = time
            self.reader.reset()
            by_time = sorted(time_map.items(), key=lambda x: x[1])
            generate = (self.reader.get_by_id(spectrum_id) for spectrum_id, _ in by_time)
            return generate
        else:
            return self.reader.iterfind("spectrum")

    def write(self):
        '''Write out the the transformed mzML file
        '''
        writer = self.writer
        with writer:
            writer.controlled_vocabularies()
            self.copy_metadata()
            with writer.run(id="transformation_run"):
                with writer.spectrum_list(len(self.reader._offset_index)):
                    self.reader.reset()
                    for i, spectrum in enumerate(self.iterspectrum()):
                        spectrum = self.transform(spectrum)
                        if spectrum is None:
                            continue
                        self.writer.write_spectrum(**self.format_spectrum(spectrum))
                        if i % 1000 == 0:
                            self.log("Handled %d spectra" % (i, ))


class MzMLToMzMLb(MzMLTransformer):
    def __init__(self, input_stream, output_stream, transform=None, transform_description=None,
                 sort_by_scan_time=False, **hdf5args):
        if transform is None:
            transform = identity
        self.input_stream = input_stream
        self.output_stream = output_stream
        self.transform = transform
        self.transform_description = transform_description
        self.sort_by_scan_time = sort_by_scan_time
        self.reader = MzMLParser(input_stream, iterative=True)
        self.writer = MzMLbWriter(output_stream, **hdf5args)
        self.psims_cv = self.writer.get_vocabulary('PSI-MS').vocabulary
